import json
from pathlib import Path

import pytest
from _pytest.tmpdir import TempdirFactory

from rasa.core.constants import (
    DEFAULT_KNOWLEDGE_BASE_ACTION,
    SLOT_LISTED_ITEMS,
    SLOT_LAST_OBJECT,
    SLOT_LAST_OBJECT_TYPE,
)
from rasa.core.domain import (
    USED_ENTITIES_KEY,
    USE_ENTITIES_KEY,
    IGNORE_ENTITIES_KEY,
)
from rasa.core import training, utils
from rasa.core.domain import Domain, InvalidDomain, SessionConfig
from rasa.core.featurizers import MaxHistoryTrackerFeaturizer
from rasa.core.slots import TextSlot, UnfeaturizedSlot
from tests.core.conftest import (
    DEFAULT_DOMAIN_PATH_WITH_SLOTS,
    DEFAULT_DOMAIN_PATH_WITH_SLOTS_AND_NO_ACTIONS,
    DEFAULT_STORIES_FILE,
)
from rasa.utils import io as io_utils


async def test_create_train_data_no_history(default_domain):
    featurizer = MaxHistoryTrackerFeaturizer(max_history=1)
    training_trackers = await training.load_data(
        DEFAULT_STORIES_FILE, default_domain, augmentation_factor=0
    )

    assert len(training_trackers) == 3
    (decoded, _) = featurizer.training_states_and_actions(
        training_trackers, default_domain
    )

    # decoded needs to be sorted
    hashed = []
    for states in decoded:
        hashed.append(json.dumps(states, sort_keys=True))
    hashed = sorted(hashed, reverse=True)

    assert hashed == [
        "[{}]",
        '[{"intent_greet": 1.0, "prev_utter_greet": 1.0}]',
        '[{"intent_greet": 1.0, "prev_action_listen": 1.0}]',
        '[{"intent_goodbye": 1.0, "prev_utter_goodbye": 1.0}]',
        '[{"intent_goodbye": 1.0, "prev_action_listen": 1.0}]',
        '[{"intent_default": 1.0, "prev_utter_default": 1.0}]',
        '[{"intent_default": 1.0, "prev_utter_default": 1.0, ' '"slot_name_0": 1.0}]',
        '[{"intent_default": 1.0, "prev_action_listen": 1.0}]',
        '[{"intent_default": 1.0, "prev_action_listen": 1.0, ' '"slot_name_0": 1.0}]',
        '[{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_utter_greet": 1.0, "slot_name_0": 1.0}]',
        '[{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_action_listen": 1.0, "slot_name_0": 1.0}]',
    ]


async def test_create_train_data_with_history(default_domain):
    featurizer = MaxHistoryTrackerFeaturizer(max_history=4)
    training_trackers = await training.load_data(
        DEFAULT_STORIES_FILE, default_domain, augmentation_factor=0
    )
    assert len(training_trackers) == 3
    (decoded, _) = featurizer.training_states_and_actions(
        training_trackers, default_domain
    )

    # decoded needs to be sorted
    hashed = []
    for states in decoded:
        hashed.append(json.dumps(states, sort_keys=True))
    hashed = sorted(hashed)

    assert hashed == [
        "[null, null, null, {}]",
        "[null, null, {}, "
        '{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_action_listen": 1.0, "slot_name_0": 1.0}]',
        "[null, null, {}, " '{"intent_greet": 1.0, "prev_action_listen": 1.0}]',
        "[null, {}, "
        '{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_action_listen": 1.0, "slot_name_0": 1.0}, '
        '{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_utter_greet": 1.0, "slot_name_0": 1.0}]',
        "[null, {}, "
        '{"intent_greet": 1.0, "prev_action_listen": 1.0}, '
        '{"intent_greet": 1.0, "prev_utter_greet": 1.0}]',
        '[{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_action_listen": 1.0, "slot_name_0": 1.0}, '
        '{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_utter_greet": 1.0, "slot_name_0": 1.0}, '
        '{"intent_default": 1.0, '
        '"prev_action_listen": 1.0, "slot_name_0": 1.0}, '
        '{"intent_default": 1.0, '
        '"prev_utter_default": 1.0, "slot_name_0": 1.0}]',
        '[{"intent_default": 1.0, "prev_action_listen": 1.0}, '
        '{"intent_default": 1.0, "prev_utter_default": 1.0}, '
        '{"intent_goodbye": 1.0, "prev_action_listen": 1.0}, '
        '{"intent_goodbye": 1.0, "prev_utter_goodbye": 1.0}]',
        '[{"intent_greet": 1.0, "prev_action_listen": 1.0}, '
        '{"intent_greet": 1.0, "prev_utter_greet": 1.0}, '
        '{"intent_default": 1.0, "prev_action_listen": 1.0}, '
        '{"intent_default": 1.0, "prev_utter_default": 1.0}]',
        '[{"intent_greet": 1.0, "prev_utter_greet": 1.0}, '
        '{"intent_default": 1.0, "prev_action_listen": 1.0}, '
        '{"intent_default": 1.0, "prev_utter_default": 1.0}, '
        '{"intent_goodbye": 1.0, "prev_action_listen": 1.0}]',
        '[{}, {"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_action_listen": 1.0, "slot_name_0": 1.0}, '
        '{"entity_name": 1.0, "intent_greet": 1.0, '
        '"prev_utter_greet": 1.0, "slot_name_0": 1.0}, '
        '{"intent_default": 1.0, '
        '"prev_action_listen": 1.0, "slot_name_0": 1.0}]',
        '[{}, {"intent_greet": 1.0, "prev_action_listen": 1.0}, '
        '{"intent_greet": 1.0, "prev_utter_greet": 1.0}, '
        '{"intent_default": 1.0, "prev_action_listen": 1.0}]',
    ]


async def test_create_train_data_unfeaturized_entities():
    domain_file = "data/test_domains/default_unfeaturized_entities.yml"
    stories_file = "data/test_stories/stories_unfeaturized_entities.md"
    domain = Domain.load(domain_file)
    featurizer = MaxHistoryTrackerFeaturizer(max_history=1)
    training_trackers = await training.load_data(
        stories_file, domain, augmentation_factor=0
    )

    assert len(training_trackers) == 2
    (decoded, _) = featurizer.training_states_and_actions(training_trackers, domain)

    # decoded needs to be sorted
    hashed = []
    for states in decoded:
        hashed.append(json.dumps(states, sort_keys=True))
    hashed = sorted(hashed, reverse=True)

    assert hashed == [
        "[{}]",
        '[{"intent_why": 1.0, "prev_utter_default": 1.0}]',
        '[{"intent_why": 1.0, "prev_action_listen": 1.0}]',
        '[{"intent_thank": 1.0, "prev_utter_default": 1.0}]',
        '[{"intent_thank": 1.0, "prev_action_listen": 1.0}]',
        '[{"intent_greet": 1.0, "prev_utter_greet": 1.0}]',
        '[{"intent_greet": 1.0, "prev_action_listen": 1.0}]',
        '[{"intent_goodbye": 1.0, "prev_utter_goodbye": 1.0}]',
        '[{"intent_goodbye": 1.0, "prev_action_listen": 1.0}]',
        '[{"entity_name": 1.0, "intent_greet": 1.0, "prev_utter_greet": 1.0}]',
        '[{"entity_name": 1.0, "intent_greet": 1.0, "prev_action_listen": 1.0}]',
        '[{"entity_name": 1.0, "entity_other": 1.0, "intent_default": 1.0, "prev_utter_default": 1.0}]',
        '[{"entity_name": 1.0, "entity_other": 1.0, "intent_default": 1.0, "prev_action_listen": 1.0}]',
        '[{"entity_name": 1.0, "entity_other": 1.0, "entity_unrelated_recognized_entity": 1.0, "intent_ask": 1.0, "prev_utter_default": 1.0}]',
        '[{"entity_name": 1.0, "entity_other": 1.0, "entity_unrelated_recognized_entity": 1.0, "intent_ask": 1.0, "prev_action_listen": 1.0}]',
    ]


def test_domain_from_template():
    domain_file = DEFAULT_DOMAIN_PATH_WITH_SLOTS
    domain = Domain.load(domain_file)

    assert not domain.is_empty()
    assert len(domain.intents) == 10
    assert len(domain.action_names) == 13


def test_avoid_action_repetition():
    domain = Domain.load(DEFAULT_DOMAIN_PATH_WITH_SLOTS)
    domain_with_no_actions = Domain.load(DEFAULT_DOMAIN_PATH_WITH_SLOTS_AND_NO_ACTIONS)

    assert not domain.is_empty() and not domain_with_no_actions.is_empty()
    assert len(domain.intents) == len(domain_with_no_actions.intents)
    assert len(domain.action_names) == len(domain_with_no_actions.action_names)


def test_utter_templates():
    domain_file = "examples/moodbot/domain.yml"
    domain = Domain.load(domain_file)
    expected_template = {
        "text": "Hey! How are you?",
        "buttons": [
            {"title": "great", "payload": "/mood_great"},
            {"title": "super sad", "payload": "/mood_unhappy"},
        ],
    }
    assert domain.random_template_for("utter_greet") == expected_template


def test_custom_slot_type(tmpdir: Path):
    domain_path = str(tmpdir / "domain.yml")
    io_utils.write_text_file(
        """
       slots:
         custom:
           type: tests.core.conftest.CustomSlot

       responses:
         utter_greet:
           - text: hey there! """,
        domain_path,
    )
    Domain.load(domain_path)


@pytest.mark.parametrize(
    "domain_unkown_slot_type",
    [
        """
    slots:
        custom:
         type: tests.core.conftest.Unknown

    responses:
        utter_greet:
         - text: hey there!""",
        """
    slots:
        custom:
         type: blubblubblub

    responses:
        utter_greet:
         - text: hey there!""",
    ],
)
def test_domain_fails_on_unknown_custom_slot_type(tmpdir, domain_unkown_slot_type):
    domain_path = str(tmpdir / "domain.yml")
    io_utils.write_text_file(domain_unkown_slot_type, domain_path)
    with pytest.raises(ValueError):
        Domain.load(domain_path)


def test_domain_to_yaml():
    test_yaml = """config:
  store_entities_as_slots: true
entities: []
forms: []
intents: []
responses:
  utter_greet:
  - text: hey there!
session_config:
  carry_over_slots_to_new_session: true
  session_expiration_time: 60
slots: {}"""

    domain = Domain.from_yaml(test_yaml)
    # python 3 and 2 are different here, python 3 will have a leading set
    # of --- at the beginning of the yml
    assert domain.as_yaml().strip().endswith(test_yaml.strip())
    assert Domain.from_yaml(domain.as_yaml()) is not None


def test_domain_to_yaml_deprecated_templates():
    test_yaml = """actions:
- utter_greet
config:
  store_entities_as_slots: true
entities: []
forms: []
intents: []
templates:
  utter_greet:
  - text: hey there!
session_config:
  carry_over_slots_to_new_session: true
  session_expiration_time: 60
slots: {}"""

    target_yaml = """actions:
- utter_greet
config:
  store_entities_as_slots: true
entities: []
forms: []
intents: []
responses:
  utter_greet:
  - text: hey there!
session_config:
  carry_over_slots_to_new_session: true
  session_expiration_time: 60
slots: {}"""

    domain = Domain.from_yaml(test_yaml)
    # python 3 and 2 are different here, python 3 will have a leading set
    # of --- at the beginning of the yml
    assert domain.as_yaml().strip().endswith(target_yaml.strip())
    assert Domain.from_yaml(domain.as_yaml()) is not None


def test_merge_yaml_domains():
    test_yaml_1 = """config:
  store_entities_as_slots: true
entities: []
intents: []
slots: {}
responses:
  utter_greet:
  - text: hey there!"""

    test_yaml_2 = """config:
  store_entities_as_slots: false
session_config:
    session_expiration_time: 20
    carry_over_slots: true
entities:
- cuisine
intents:
- greet
slots:
  cuisine:
    type: text
responses:
  utter_goodbye:
  - text: bye!
  utter_greet:
  - text: hey you!"""

    domain_1 = Domain.from_yaml(test_yaml_1)
    domain_2 = Domain.from_yaml(test_yaml_2)
    domain = domain_1.merge(domain_2)
    # single attribute should be taken from domain_1
    assert domain.store_entities_as_slots
    # conflicts should be taken from domain_1
    assert domain.templates == {
        "utter_greet": [{"text": "hey there!"}],
        "utter_goodbye": [{"text": "bye!"}],
    }
    # lists should be deduplicated and merged
    assert domain.intents == ["greet"]
    assert domain.entities == ["cuisine"]
    assert isinstance(domain.slots[0], TextSlot)
    assert domain.slots[0].name == "cuisine"
    assert sorted(domain.user_actions) == sorted(["utter_greet", "utter_goodbye"])
    assert domain.session_config == SessionConfig(20, True)

    domain = domain_1.merge(domain_2, override=True)
    # single attribute should be taken from domain_2
    assert not domain.store_entities_as_slots
    # conflicts should take value from domain_2
    assert domain.templates == {
        "utter_greet": [{"text": "hey you!"}],
        "utter_goodbye": [{"text": "bye!"}],
    }
    assert domain.session_config == SessionConfig(20, True)


def test_merge_session_config_if_first_is_not_default():
    yaml1 = """
session_config:
    session_expiration_time: 20
    carry_over_slots: true"""

    yaml2 = """
 session_config:
    session_expiration_time: 40
    carry_over_slots: true
    """

    domain1 = Domain.from_yaml(yaml1)
    domain2 = Domain.from_yaml(yaml2)

    merged = domain1.merge(domain2)
    assert merged.session_config == SessionConfig(20, True)

    merged = domain1.merge(domain2, override=True)
    assert merged.session_config == SessionConfig(40, True)


@pytest.mark.parametrize(
    "intents, entities, intent_properties",
    [
        (
            ["greet", "goodbye"],
            ["entity", "other", "third"],
            {
                "greet": {USED_ENTITIES_KEY: ["entity", "other", "third"]},
                "goodbye": {USED_ENTITIES_KEY: ["entity", "other", "third"]},
            },
        ),
        (
            [{"greet": {USE_ENTITIES_KEY: []}}, "goodbye"],
            ["entity", "other", "third"],
            {
                "greet": {USED_ENTITIES_KEY: []},
                "goodbye": {USED_ENTITIES_KEY: ["entity", "other", "third"]},
            },
        ),
        (
            [
                {
                    "greet": {
                        "triggers": "utter_goodbye",
                        USE_ENTITIES_KEY: ["entity"],
                        IGNORE_ENTITIES_KEY: ["other"],
                    }
                },
                "goodbye",
            ],
            ["entity", "other", "third"],
            {
                "greet": {"triggers": "utter_goodbye", USED_ENTITIES_KEY: ["entity"],},
                "goodbye": {USED_ENTITIES_KEY: ["entity", "other", "third"]},
            },
        ),
        (
            [
                {"greet": {"triggers": "utter_goodbye", USE_ENTITIES_KEY: None}},
                {"goodbye": {USE_ENTITIES_KEY: [], IGNORE_ENTITIES_KEY: []}},
            ],
            ["entity", "other", "third"],
            {
                "greet": {USED_ENTITIES_KEY: [], "triggers": "utter_goodbye",},
                "goodbye": {USED_ENTITIES_KEY: []},
            },
        ),
    ],
)
def test_collect_intent_properties(intents, entities, intent_properties):
    assert Domain.collect_intent_properties(intents, entities) == intent_properties


def test_load_domain_from_directory_tree(tmpdir_factory: TempdirFactory):
    root = tmpdir_factory.mktemp("Parent Bot")
    root_domain = {"actions": ["utter_root", "utter_root2"]}
    utils.dump_obj_as_yaml_to_file(root / "domain.yml", root_domain)

    subdirectory_1 = root / "Skill 1"
    subdirectory_1.mkdir()
    skill_1_domain = {"actions": ["utter_skill_1"]}
    utils.dump_obj_as_yaml_to_file(subdirectory_1 / "domain.yml", skill_1_domain)

    subdirectory_2 = root / "Skill 2"
    subdirectory_2.mkdir()
    skill_2_domain = {"actions": ["utter_skill_2"]}
    utils.dump_obj_as_yaml_to_file(subdirectory_2 / "domain.yml", skill_2_domain)

    subsubdirectory = subdirectory_2 / "Skill 2-1"
    subsubdirectory.mkdir()
    skill_2_1_domain = {"actions": ["utter_subskill", "utter_root"]}
    # Check if loading from `.yaml` also works
    utils.dump_obj_as_yaml_to_file(subsubdirectory / "domain.yaml", skill_2_1_domain)

    subsubdirectory_2 = subdirectory_2 / "Skill 2-2"
    subsubdirectory_2.mkdir()
    excluded_domain = {"actions": ["should not be loaded"]}
    utils.dump_obj_as_yaml_to_file(
        subsubdirectory_2 / "other_name.yaml", excluded_domain
    )

    actual = Domain.load(str(root))
    expected = [
        "utter_root",
        "utter_root2",
        "utter_skill_1",
        "utter_skill_2",
        "utter_subskill",
    ]

    assert set(actual.user_actions) == set(expected)


def test_domain_warnings():
    domain = Domain.load(DEFAULT_DOMAIN_PATH_WITH_SLOTS)

    warning_types = [
        "action_warnings",
        "intent_warnings",
        "entity_warnings",
        "slot_warnings",
    ]

    actions = ["action_1", "action_2"]
    intents = ["intent_1", "intent_2"]
    entities = ["entity_1", "entity_2"]
    slots = ["slot_1", "slot_2"]
    domain_warnings = domain.domain_warnings(
        intents=intents, entities=entities, actions=actions, slots=slots
    )

    # elements not found in domain should be in `in_training_data` diff
    for _type, elements in zip(warning_types, [actions, intents, entities]):
        assert set(domain_warnings[_type]["in_training_data"]) == set(elements)

    # all other domain elements should be in `in_domain` diff
    for _type, elements in zip(
        warning_types, [domain.user_actions, domain.intents, domain.entities]
    ):
        assert set(domain_warnings[_type]["in_domain"]) == set(elements)

    # fully aligned domain and elements should yield empty diff
    domain_warnings = domain.domain_warnings(
        intents=domain.intents,
        entities=domain.entities,
        actions=domain.user_actions,
        slots=[s.name for s in domain.slots],
    )

    for diff_dict in domain_warnings.values():
        assert all(not diff_set for diff_set in diff_dict.values())


def test_unfeaturized_slot_in_domain_warnings():
    # create empty domain
    domain = Domain.empty()

    # add one unfeaturized and one text slot
    unfeaturized_slot = UnfeaturizedSlot("unfeaturized_slot", "value1")
    text_slot = TextSlot("text_slot", "value2")
    domain.slots.extend([unfeaturized_slot, text_slot])

    # ensure both are in domain
    assert all(slot in domain.slots for slot in (unfeaturized_slot, text_slot))

    # text slot should appear in domain warnings, unfeaturized slot should not
    in_domain_slot_warnings = domain.domain_warnings()["slot_warnings"]["in_domain"]
    assert text_slot.name in in_domain_slot_warnings
    assert unfeaturized_slot.name not in in_domain_slot_warnings


def test_check_domain_sanity_on_invalid_domain():
    with pytest.raises(InvalidDomain):
        Domain(
            intents={},
            entities=[],
            slots=[],
            templates={},
            action_names=["random_name", "random_name"],
            form_names=[],
        )

    with pytest.raises(InvalidDomain):
        Domain(
            intents={},
            entities=[],
            slots=[TextSlot("random_name"), TextSlot("random_name")],
            templates={},
            action_names=[],
            form_names=[],
        )

    with pytest.raises(InvalidDomain):
        Domain(
            intents={},
            entities=["random_name", "random_name", "other_name", "other_name"],
            slots=[],
            templates={},
            action_names=[],
            form_names=[],
        )

    with pytest.raises(InvalidDomain):
        Domain(
            intents={},
            entities=[],
            slots=[],
            templates={},
            action_names=[],
            form_names=["random_name", "random_name"],
        )


def test_load_on_invalid_domain():
    with pytest.raises(InvalidDomain):
        Domain.load("data/test_domains/duplicate_intents.yml")

    with pytest.raises(InvalidDomain):
        Domain.load("data/test_domains/duplicate_actions.yml")

    with pytest.raises(InvalidDomain):
        Domain.load("data/test_domains/duplicate_templates.yml")

    with pytest.raises(InvalidDomain):
        Domain.load("data/test_domains/duplicate_entities.yml")

    # Currently just deprecated
    # with pytest.raises(InvalidDomain):
    #     Domain.load("data/test_domains/missing_text_for_templates.yml")


def test_is_empty():
    assert Domain.empty().is_empty()


def test_transform_intents_for_file_default():
    domain_path = "data/test_domains/default_unfeaturized_entities.yml"
    domain = Domain.load(domain_path)
    transformed = domain._transform_intents_for_file()

    expected = [
        {"greet": {USE_ENTITIES_KEY: ["name"]}},
        {"default": {IGNORE_ENTITIES_KEY: ["unrelated_recognized_entity"]}},
        {"goodbye": {USE_ENTITIES_KEY: []}},
        {"thank": {USE_ENTITIES_KEY: []}},
        {"ask": {USE_ENTITIES_KEY: True}},
        {"why": {USE_ENTITIES_KEY: []}},
        {"pure_intent": {USE_ENTITIES_KEY: True}},
    ]

    assert transformed == expected


def test_transform_intents_for_file_with_mapping():
    domain_path = "data/test_domains/default_with_mapping.yml"
    domain = Domain.load(domain_path)
    transformed = domain._transform_intents_for_file()

    expected = [
        {"greet": {"triggers": "utter_greet", USE_ENTITIES_KEY: True}},
        {"default": {"triggers": "utter_default", USE_ENTITIES_KEY: True}},
        {"goodbye": {USE_ENTITIES_KEY: True}},
    ]

    assert transformed == expected


def test_clean_domain_for_file():
    domain_path = "data/test_domains/default_unfeaturized_entities.yml"
    cleaned = Domain.load(domain_path).cleaned_domain()

    expected = {
        "intents": [
            {"greet": {USE_ENTITIES_KEY: ["name"]}},
            {"default": {IGNORE_ENTITIES_KEY: ["unrelated_recognized_entity"]}},
            {"goodbye": {USE_ENTITIES_KEY: []}},
            {"thank": {USE_ENTITIES_KEY: []}},
            "ask",
            {"why": {USE_ENTITIES_KEY: []}},
            "pure_intent",
        ],
        "entities": ["name", "other", "unrelated_recognized_entity"],
        "responses": {
            "utter_greet": [{"text": "hey there!"}],
            "utter_goodbye": [{"text": "goodbye :("}],
            "utter_default": [{"text": "default message"}],
        },
        "actions": ["utter_default", "utter_goodbye", "utter_greet"],
        "session_config": {
            "carry_over_slots_to_new_session": True,
            "session_expiration_time": 0,
        },
    }

    assert cleaned == expected


def test_clean_domain_deprecated_templates():
    domain_path = "data/test_domains/default_deprecated_templates.yml"
    cleaned = Domain.load(domain_path).cleaned_domain()

    expected = {
        "intents": [
            {"greet": {USE_ENTITIES_KEY: ["name"]}},
            {"default": {IGNORE_ENTITIES_KEY: ["unrelated_recognized_entity"]}},
            {"goodbye": {USE_ENTITIES_KEY: []}},
            {"thank": {USE_ENTITIES_KEY: []}},
            "ask",
            {"why": {USE_ENTITIES_KEY: []}},
            "pure_intent",
        ],
        "entities": ["name", "other", "unrelated_recognized_entity"],
        "responses": {
            "utter_greet": [{"text": "hey there!"}],
            "utter_goodbye": [{"text": "goodbye :("}],
            "utter_default": [{"text": "default message"}],
        },
        "actions": ["utter_default", "utter_goodbye", "utter_greet"],
    }

    expected = Domain.from_dict(expected)
    actual = Domain.from_dict(cleaned)

    assert hash(actual) == hash(expected)


def test_add_knowledge_base_slots(default_domain):
    import copy

    # don't modify default domain as it is used in other tests
    test_domain = copy.deepcopy(default_domain)

    test_domain.action_names.append(DEFAULT_KNOWLEDGE_BASE_ACTION)

    slot_names = [s.name for s in test_domain.slots]

    assert SLOT_LISTED_ITEMS not in slot_names
    assert SLOT_LAST_OBJECT not in slot_names
    assert SLOT_LAST_OBJECT_TYPE not in slot_names

    test_domain.add_knowledge_base_slots()

    slot_names = [s.name for s in test_domain.slots]

    assert SLOT_LISTED_ITEMS in slot_names
    assert SLOT_LAST_OBJECT in slot_names
    assert SLOT_LAST_OBJECT_TYPE in slot_names


@pytest.mark.parametrize(
    "input_domain, expected_session_expiration_time, expected_carry_over_slots",
    [
        (
            """session_config:
    session_expiration_time: 0
    carry_over_slots_to_new_session: true""",
            0,
            True,
        ),
        ("", 0, True),
        (
            """session_config:
    carry_over_slots_to_new_session: false""",
            0,
            False,
        ),
        (
            """session_config:
    session_expiration_time: 20.2
    carry_over_slots_to_new_session: False""",
            20.2,
            False,
        ),
        ("""session_config: {}""", 0, True),
    ],
)
def test_session_config(
    input_domain,
    expected_session_expiration_time: float,
    expected_carry_over_slots: bool,
):
    domain = Domain.from_yaml(input_domain)
    assert (
        domain.session_config.session_expiration_time
        == expected_session_expiration_time
    )
    assert domain.session_config.carry_over_slots == expected_carry_over_slots


def test_domain_as_dict_with_session_config():
    session_config = SessionConfig(123, False)
    domain = Domain.empty()
    domain.session_config = session_config

    serialized = domain.as_dict()
    deserialized = Domain.from_dict(serialized)

    assert deserialized.session_config == session_config


@pytest.mark.parametrize(
    "session_config, enabled",
    [
        (SessionConfig(0, True), False),
        (SessionConfig(1, True), True),
        (SessionConfig(-1, False), False),
    ],
)
def test_are_sessions_enabled(session_config: SessionConfig, enabled: bool):
    assert session_config.are_sessions_enabled() == enabled


def test_domain_utterance_actions_deprecated_templates():
    new_yaml = """actions:
- utter_greet
- utter_goodbye
config:
  store_entities_as_slots: true
entities: []
forms: []
intents: []
templates:
  utter_greet:
  - text: hey there!
  utter_goodbye:
  - text: bye!
session_config:
  carry_over_slots_to_new_session: true
  session_expiration_time: 60
slots: {}"""

    old_yaml = """config:
  store_entities_as_slots: true
entities: []
forms: []
intents: []
responses:
  utter_greet:
  - text: hey there!
  utter_goodbye:
  - text: bye!
session_config:
  carry_over_slots_to_new_session: true
  session_expiration_time: 60
slots: {}"""

    old_domain = Domain.from_yaml(old_yaml)
    new_domain = Domain.from_yaml(new_yaml)
    assert hash(old_domain) == hash(new_domain)



from unittest.mock import Mock, patch

import numpy as np
import pytest

from rasa.core import training
from rasa.core.actions.action import (
    ACTION_DEFAULT_ASK_AFFIRMATION_NAME,
    ACTION_DEFAULT_ASK_REPHRASE_NAME,
    ACTION_DEFAULT_FALLBACK_NAME,
    ACTION_LISTEN_NAME,
    ActionRevertFallbackEvents,
    ACTION_RESTART_NAME,
    ACTION_BACK_NAME,
)
from rasa.core.constants import USER_INTENT_RESTART, USER_INTENT_BACK
from rasa.core.channels.channel import UserMessage
from rasa.core.domain import Domain
from rasa.core.events import ActionExecuted, ConversationPaused
from rasa.core.featurizers import (
    BinarySingleStateFeaturizer,
    LabelTokenizerSingleStateFeaturizer,
    MaxHistoryTrackerFeaturizer,
    FullDialogueTrackerFeaturizer,
)
from rasa.core.policies.two_stage_fallback import TwoStageFallbackPolicy
from rasa.core.policies.ted_policy import TEDPolicy
from rasa.core.policies.fallback import FallbackPolicy
from rasa.core.policies.form_policy import FormPolicy
from rasa.core.policies.keras_policy import KerasPolicy
from rasa.core.policies.mapping_policy import MappingPolicy
from rasa.core.policies.memoization import AugmentedMemoizationPolicy, MemoizationPolicy
from rasa.core.policies.sklearn_policy import SklearnPolicy
from rasa.core.trackers import DialogueStateTracker
from rasa.utils.tensorflow.constants import (
    SIMILARITY_TYPE,
    RANKING_LENGTH,
    LOSS_TYPE,
    SCALE_LOSS,
    EVAL_NUM_EXAMPLES,
    EPOCHS,
    KEY_RELATIVE_ATTENTION,
    VALUE_RELATIVE_ATTENTION,
    MAX_RELATIVE_POSITION,
)
from rasa.utils import train_utils
from tests.core.conftest import (
    DEFAULT_DOMAIN_PATH_WITH_MAPPING,
    DEFAULT_DOMAIN_PATH_WITH_SLOTS,
    DEFAULT_STORIES_FILE,
)
from tests.core.utilities import get_tracker, read_dialogue_file, user_uttered


async def train_trackers(domain, augmentation_factor=20):
    return await training.load_data(
        DEFAULT_STORIES_FILE, domain, augmentation_factor=augmentation_factor
    )


# We are going to use class style testing here since unfortunately pytest
# doesn't support using fixtures as arguments to its own parameterize yet
# (hence, we can't train a policy, declare it as a fixture and use the
# different fixtures of the different policies for the functional tests).
# Therefore, we are going to reverse this and train the policy within a class
# and collect the tests in a base class.
# noinspection PyMethodMayBeStatic
class PolicyTestCollection:
    """Tests every policy needs to fulfill.

    Each policy can declare further tests on its own."""

    max_history = 3  # this is the amount of history we test on

    def create_policy(self, featurizer, priority):
        raise NotImplementedError

    @pytest.fixture(scope="module")
    def featurizer(self):
        featurizer = MaxHistoryTrackerFeaturizer(
            BinarySingleStateFeaturizer(), max_history=self.max_history
        )
        return featurizer

    @pytest.fixture(scope="module")
    def priority(self):
        return 1

    @pytest.fixture(scope="module")
    def default_domain(self):
        return Domain.load(DEFAULT_DOMAIN_PATH_WITH_SLOTS)

    @pytest.fixture(scope="module")
    def tracker(self, default_domain):
        return DialogueStateTracker(UserMessage.DEFAULT_SENDER_ID, default_domain.slots)

    @pytest.fixture(scope="module")
    async def trained_policy(self, featurizer, priority):
        default_domain = Domain.load(DEFAULT_DOMAIN_PATH_WITH_SLOTS)
        policy = self.create_policy(featurizer, priority)
        training_trackers = await train_trackers(default_domain, augmentation_factor=20)
        policy.train(training_trackers, default_domain)
        return policy

    def test_featurizer(self, trained_policy, tmpdir):
        assert isinstance(trained_policy.featurizer, MaxHistoryTrackerFeaturizer)
        assert trained_policy.featurizer.max_history == self.max_history
        assert isinstance(
            trained_policy.featurizer.state_featurizer, BinarySingleStateFeaturizer
        )
        trained_policy.persist(tmpdir.strpath)
        loaded = trained_policy.__class__.load(tmpdir.strpath)
        assert isinstance(loaded.featurizer, MaxHistoryTrackerFeaturizer)
        assert loaded.featurizer.max_history == self.max_history
        assert isinstance(
            loaded.featurizer.state_featurizer, BinarySingleStateFeaturizer
        )

    async def test_persist_and_load(self, trained_policy, default_domain, tmpdir):
        trained_policy.persist(tmpdir.strpath)
        loaded = trained_policy.__class__.load(tmpdir.strpath)
        trackers = await train_trackers(default_domain, augmentation_factor=20)

        for tracker in trackers:
            predicted_probabilities = loaded.predict_action_probabilities(
                tracker, default_domain
            )
            actual_probabilities = trained_policy.predict_action_probabilities(
                tracker, default_domain
            )
            assert predicted_probabilities == actual_probabilities

    def test_prediction_on_empty_tracker(self, trained_policy, default_domain):
        tracker = DialogueStateTracker(
            UserMessage.DEFAULT_SENDER_ID, default_domain.slots
        )
        probabilities = trained_policy.predict_action_probabilities(
            tracker, default_domain
        )
        assert len(probabilities) == default_domain.num_actions
        assert max(probabilities) <= 1.0
        assert min(probabilities) >= 0.0

    @pytest.mark.filterwarnings(
        "ignore:.*without a trained model present.*:UserWarning"
    )
    def test_persist_and_load_empty_policy(self, tmpdir):
        empty_policy = self.create_policy(None, None)
        empty_policy.persist(tmpdir.strpath)
        loaded = empty_policy.__class__.load(tmpdir.strpath)
        assert loaded is not None

    @staticmethod
    def _get_next_action(policy, events, domain):
        tracker = get_tracker(events)

        scores = policy.predict_action_probabilities(tracker, domain)
        index = scores.index(max(scores))
        return domain.action_names[index]


class TestKerasPolicy(PolicyTestCollection):
    def create_policy(self, featurizer, priority):
        p = KerasPolicy(featurizer, priority)
        return p


class TestSklearnPolicy(PolicyTestCollection):
    def create_policy(self, featurizer, priority, **kwargs):
        p = SklearnPolicy(featurizer, priority, **kwargs)
        return p

    @pytest.yield_fixture
    def mock_search(self):
        with patch("rasa.core.policies.sklearn_policy.GridSearchCV") as gs:
            gs.best_estimator_ = "mockmodel"
            gs.best_score_ = 0.123
            gs.return_value = gs  # for __init__
            yield gs

    @pytest.fixture(scope="module")
    async def trackers(self, default_domain):
        return await train_trackers(default_domain, augmentation_factor=20)

    def test_additional_train_args_do_not_raise(
        self, default_domain, trackers, featurizer, priority
    ):
        policy = self.create_policy(featurizer=featurizer, priority=priority, cv=None)
        policy.train(trackers, domain=default_domain, this_is_not_a_feature=True)

    def test_cv_none_does_not_trigger_search(
        self, mock_search, default_domain, trackers, featurizer, priority
    ):
        policy = self.create_policy(featurizer=featurizer, priority=priority, cv=None)
        policy.train(trackers, domain=default_domain)

        assert mock_search.call_count == 0
        assert policy.model != "mockmodel"

    def test_cv_not_none_param_grid_none_triggers_search_without_params(
        self, mock_search, default_domain, trackers, featurizer, priority
    ):

        policy = self.create_policy(featurizer=featurizer, priority=priority, cv=3)
        policy.train(trackers, domain=default_domain)

        assert mock_search.call_count > 0
        assert mock_search.call_args_list[0][1]["cv"] == 3
        assert mock_search.call_args_list[0][1]["param_grid"] == {}
        assert policy.model == "mockmodel"

    def test_cv_not_none_param_grid_none_triggers_search_with_params(
        self, mock_search, default_domain, trackers, featurizer, priority
    ):
        param_grid = {"n_estimators": 50}
        policy = self.create_policy(
            featurizer=featurizer, priority=priority, cv=3, param_grid=param_grid
        )
        policy.train(trackers, domain=default_domain)

        assert mock_search.call_count > 0
        assert mock_search.call_args_list[0][1]["cv"] == 3
        assert mock_search.call_args_list[0][1]["param_grid"] == param_grid
        assert policy.model == "mockmodel"

    def test_missing_classes_filled_correctly(
        self, default_domain, trackers, tracker, featurizer, priority
    ):
        # Pretend that a couple of classes are missing and check that
        # those classes are predicted as 0, while the other class
        # probabilities are predicted normally.
        policy = self.create_policy(featurizer=featurizer, priority=priority, cv=None)

        classes = [1, 3]
        new_trackers = []
        for tr in trackers:
            new_tracker = DialogueStateTracker(
                UserMessage.DEFAULT_SENDER_ID, default_domain.slots
            )
            for e in tr.applied_events():
                if isinstance(e, ActionExecuted):
                    new_action = default_domain.action_for_index(
                        np.random.choice(classes), action_endpoint=None
                    ).name()
                    new_tracker.update(ActionExecuted(new_action))
                else:
                    new_tracker.update(e)

            new_trackers.append(new_tracker)

        policy.train(new_trackers, domain=default_domain)
        predicted_probabilities = policy.predict_action_probabilities(
            tracker, default_domain
        )

        assert len(predicted_probabilities) == default_domain.num_actions
        assert np.allclose(sum(predicted_probabilities), 1.0)
        for i, prob in enumerate(predicted_probabilities):
            if i in classes:
                assert prob >= 0.0
            else:
                assert prob == 0.0

    def test_train_kwargs_are_set_on_model(
        self, default_domain, trackers, featurizer, priority
    ):
        policy = self.create_policy(
            featurizer=featurizer, priority=priority, cv=None, C=123
        )
        policy.train(trackers, domain=default_domain)
        assert policy.model.C == 123

    def test_train_with_shuffle_false(
        self, default_domain, trackers, featurizer, priority
    ):
        policy = self.create_policy(
            featurizer=featurizer, priority=priority, shuffle=False
        )
        # does not raise
        policy.train(trackers, domain=default_domain)


class TestTEDPolicy(PolicyTestCollection):
    def create_policy(self, featurizer, priority):
        p = TEDPolicy(featurizer=featurizer, priority=priority)
        return p

    def test_similarity_type(self, trained_policy):
        assert trained_policy.config[SIMILARITY_TYPE] == "inner"

    def test_ranking_length(self, trained_policy):
        assert trained_policy.config[RANKING_LENGTH] == 10

    def test_normalization(self, trained_policy, tracker, default_domain, monkeypatch):
        # first check the output is what we expect
        predicted_probabilities = trained_policy.predict_action_probabilities(
            tracker, default_domain
        )
        # count number of non-zero confidences
        assert (
            sum([confidence > 0 for confidence in predicted_probabilities])
            == trained_policy.config[RANKING_LENGTH]
        )
        # check that the norm is still 1
        assert sum(predicted_probabilities) == pytest.approx(1)

        # also check our function is called
        mock = Mock()
        monkeypatch.setattr(train_utils, "normalize", mock.normalize)
        trained_policy.predict_action_probabilities(tracker, default_domain)

        mock.normalize.assert_called_once()

    async def test_gen_batch(self, trained_policy, default_domain):
        training_trackers = await train_trackers(default_domain, augmentation_factor=0)
        training_data = trained_policy.featurize_for_training(
            training_trackers, default_domain
        )
        model_data = trained_policy._create_model_data(training_data.X, training_data.y)
        batch_size = 2
        batch_x, batch_y, _ = next(model_data._gen_batch(batch_size=batch_size))
        assert batch_x.shape[0] == batch_size and batch_y.shape[0] == batch_size
        assert (
            batch_x[0].shape == model_data.get("dialogue_features")[0][0].shape
            and batch_y[0].shape == model_data.get("label_features")[0][0].shape
        )
        batch_x, batch_y, _ = next(
            model_data._gen_batch(
                batch_size=batch_size, batch_strategy="balanced", shuffle=True
            )
        )
        assert batch_x.shape[0] == batch_size and batch_y.shape[0] == batch_size
        assert (
            batch_x[0].shape == model_data.get("dialogue_features")[0][0].shape
            and batch_y[0].shape == model_data.get("label_features")[0][0].shape
        )


class TestTEDPolicyMargin(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        p = TEDPolicy(featurizer=featurizer, priority=priority, **{LOSS_TYPE: "margin"})
        return p

    def test_similarity_type(self, trained_policy):
        assert trained_policy.config[SIMILARITY_TYPE] == "cosine"

    def test_normalization(self, trained_policy, tracker, default_domain, monkeypatch):
        # Mock actual normalization method
        mock = Mock()
        monkeypatch.setattr(train_utils, "normalize", mock.normalize)
        trained_policy.predict_action_probabilities(tracker, default_domain)

        # function should not get called for margin loss_type
        mock.normalize.assert_not_called()


class TestTEDPolicyWithEval(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        p = TEDPolicy(
            featurizer=featurizer,
            priority=priority,
            **{SCALE_LOSS: False, EVAL_NUM_EXAMPLES: 4},
        )
        return p


class TestTEDPolicyNoNormalization(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        p = TEDPolicy(featurizer=featurizer, priority=priority, **{RANKING_LENGTH: 0})
        return p

    def test_ranking_length(self, trained_policy):
        assert trained_policy.config[RANKING_LENGTH] == 0

    def test_normalization(self, trained_policy, tracker, default_domain, monkeypatch):
        # first check the output is what we expect
        predicted_probabilities = trained_policy.predict_action_probabilities(
            tracker, default_domain
        )
        # there should be no normalization
        assert all([confidence > 0 for confidence in predicted_probabilities])

        # also check our function is not called
        mock = Mock()
        monkeypatch.setattr(train_utils, "normalize", mock.normalize)
        trained_policy.predict_action_probabilities(tracker, default_domain)

        mock.normalize.assert_not_called()


class TestTEDPolicyLowRankingLength(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        p = TEDPolicy(featurizer=featurizer, priority=priority, **{RANKING_LENGTH: 3})
        return p

    def test_ranking_length(self, trained_policy):
        assert trained_policy.config[RANKING_LENGTH] == 3


class TestTEDPolicyHighRankingLength(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        p = TEDPolicy(featurizer=featurizer, priority=priority, **{RANKING_LENGTH: 11})
        return p

    def test_ranking_length(self, trained_policy):
        assert trained_policy.config[RANKING_LENGTH] == 11


class TestTEDPolicyWithFullDialogue(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        # use standard featurizer from TEDPolicy,
        # since it is using FullDialogueTrackerFeaturizer
        # if max_history is not specified
        p = TEDPolicy(priority=priority)
        return p

    def test_featurizer(self, trained_policy, tmpdir):
        assert isinstance(trained_policy.featurizer, FullDialogueTrackerFeaturizer)
        assert isinstance(
            trained_policy.featurizer.state_featurizer,
            LabelTokenizerSingleStateFeaturizer,
        )
        trained_policy.persist(tmpdir.strpath)
        loaded = trained_policy.__class__.load(tmpdir.strpath)
        assert isinstance(loaded.featurizer, FullDialogueTrackerFeaturizer)
        assert isinstance(
            loaded.featurizer.state_featurizer, LabelTokenizerSingleStateFeaturizer
        )


class TestTEDPolicyWithMaxHistory(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        # use standard featurizer from TEDPolicy,
        # since it is using MaxHistoryTrackerFeaturizer
        # if max_history is specified
        p = TEDPolicy(priority=priority, max_history=self.max_history)
        return p

    def test_featurizer(self, trained_policy, tmpdir):
        assert isinstance(trained_policy.featurizer, MaxHistoryTrackerFeaturizer)
        assert trained_policy.featurizer.max_history == self.max_history
        assert isinstance(
            trained_policy.featurizer.state_featurizer,
            LabelTokenizerSingleStateFeaturizer,
        )
        trained_policy.persist(tmpdir.strpath)
        loaded = trained_policy.__class__.load(tmpdir.strpath)
        assert isinstance(loaded.featurizer, MaxHistoryTrackerFeaturizer)
        assert loaded.featurizer.max_history == self.max_history
        assert isinstance(
            loaded.featurizer.state_featurizer, LabelTokenizerSingleStateFeaturizer
        )


class TestTEDPolicyWithRelativeAttention(TestTEDPolicy):
    def create_policy(self, featurizer, priority):
        p = TEDPolicy(
            featurizer=featurizer,
            priority=priority,
            **{
                KEY_RELATIVE_ATTENTION: True,
                VALUE_RELATIVE_ATTENTION: True,
                MAX_RELATIVE_POSITION: 5,
            },
        )
        return p


class TestMemoizationPolicy(PolicyTestCollection):
    def create_policy(self, featurizer, priority):
        max_history = None
        if isinstance(featurizer, MaxHistoryTrackerFeaturizer):
            max_history = featurizer.max_history
        p = MemoizationPolicy(priority=priority, max_history=max_history)
        return p

    def test_featurizer(self, trained_policy, tmpdir):
        assert isinstance(trained_policy.featurizer, MaxHistoryTrackerFeaturizer)
        assert trained_policy.featurizer.state_featurizer is None
        trained_policy.persist(tmpdir.strpath)
        loaded = trained_policy.__class__.load(tmpdir.strpath)
        assert isinstance(loaded.featurizer, MaxHistoryTrackerFeaturizer)
        assert loaded.featurizer.state_featurizer is None

    async def test_memorise(self, trained_policy, default_domain):
        trackers = await train_trackers(default_domain, augmentation_factor=20)
        trained_policy.train(trackers, default_domain)
        lookup_with_augmentation = trained_policy.lookup

        trackers = [
            t for t in trackers if not hasattr(t, "is_augmented") or not t.is_augmented
        ]

        (
            all_states,
            all_actions,
        ) = trained_policy.featurizer.training_states_and_actions(
            trackers, default_domain
        )

        for tracker, states, actions in zip(trackers, all_states, all_actions):
            recalled = trained_policy.recall(states, tracker, default_domain)
            assert recalled == default_domain.index_for_action(actions[0])

        nums = np.random.randn(default_domain.num_states)
        random_states = [{f: num for f, num in zip(default_domain.input_states, nums)}]
        assert trained_policy._recall_states(random_states) is None

        # compare augmentation for augmentation_factor of 0 and 20:
        trackers_no_augmentation = await train_trackers(
            default_domain, augmentation_factor=0
        )
        trained_policy.train(trackers_no_augmentation, default_domain)
        lookup_no_augmentation = trained_policy.lookup

        assert lookup_no_augmentation == lookup_with_augmentation

    def test_memorise_with_nlu(self, trained_policy, default_domain):
        filename = "data/test_dialogues/default.json"
        dialogue = read_dialogue_file(filename)

        tracker = DialogueStateTracker(dialogue.name, default_domain.slots)
        tracker.recreate_from_dialogue(dialogue)
        states = trained_policy.featurizer.prediction_states([tracker], default_domain)[
            0
        ]

        recalled = trained_policy.recall(states, tracker, default_domain)
        assert recalled is not None


class TestAugmentedMemoizationPolicy(TestMemoizationPolicy):
    def create_policy(self, featurizer, priority):
        max_history = None
        if isinstance(featurizer, MaxHistoryTrackerFeaturizer):
            max_history = featurizer.max_history
        p = AugmentedMemoizationPolicy(priority=priority, max_history=max_history)
        return p


class TestFormPolicy(TestMemoizationPolicy):
    def create_policy(self, featurizer, priority):
        p = FormPolicy(priority=priority)
        return p

    async def test_memorise(self, trained_policy, default_domain):
        domain = Domain.load("data/test_domains/form.yml")
        trackers = await training.load_data("data/test_stories/stories_form.md", domain)
        trained_policy.train(trackers, domain)

        (
            all_states,
            all_actions,
        ) = trained_policy.featurizer.training_states_and_actions(trackers, domain)

        for tracker, states, actions in zip(trackers, all_states, all_actions):
            for state in states:
                if state is not None:
                    # check that 'form: inform' was ignored
                    assert "intent_inform" not in state.keys()
            recalled = trained_policy.recall(states, tracker, domain)
            active_form = trained_policy._get_active_form_name(states[-1])

            if states[0] is not None and states[-1] is not None:
                # explicitly set intents and actions before listen after
                # which FormPolicy should not predict a form action and
                # should add FormValidation(False) event
                # @formatter:off
                is_no_validation = (
                    (
                        "prev_some_form" in states[0].keys()
                        and "intent_default" in states[-1].keys()
                    )
                    or (
                        "prev_some_form" in states[0].keys()
                        and "intent_stop" in states[-1].keys()
                    )
                    or (
                        "prev_utter_ask_continue" in states[0].keys()
                        and "intent_affirm" in states[-1].keys()
                    )
                    or (
                        "prev_utter_ask_continue" in states[0].keys()
                        and "intent_deny" in states[-1].keys()
                    )
                    # comes from the fact that intent_inform after utter_ask_continue
                    # is not read from stories
                    or (
                        "prev_utter_ask_continue" in states[0].keys()
                        and "intent_stop" in states[-1].keys()
                    )
                )
                # @formatter:on
            else:
                is_no_validation = False

            if "intent_start_form" in states[-1]:
                # explicitly check that intent that starts the form
                # is not memorized as non validation intent
                assert recalled is None
            elif is_no_validation:
                assert recalled == active_form
            else:
                assert recalled is None

        nums = np.random.randn(domain.num_states)
        random_states = [{f: num for f, num in zip(domain.input_states, nums)}]
        assert trained_policy.recall(random_states, None, domain) is None

    def test_memorise_with_nlu(self, trained_policy, default_domain):
        pass


class TestMappingPolicy(PolicyTestCollection):
    def create_policy(self, featurizer, priority):
        p = MappingPolicy()
        return p

    def test_featurizer(self, trained_policy, tmpdir):
        assert trained_policy.featurizer is None
        trained_policy.persist(tmpdir.strpath)
        loaded = trained_policy.__class__.load(tmpdir.strpath)
        assert loaded.featurizer is None

    @pytest.fixture(scope="module")
    def domain_with_mapping(self):
        return Domain.load(DEFAULT_DOMAIN_PATH_WITH_MAPPING)

    @pytest.fixture
    def tracker(self, domain_with_mapping):
        return DialogueStateTracker(
            UserMessage.DEFAULT_SENDER_ID, domain_with_mapping.slots
        )

    @pytest.fixture(
        params=[
            ("default", "utter_default"),
            ("greet", "utter_greet"),
            (USER_INTENT_RESTART, ACTION_RESTART_NAME),
            (USER_INTENT_BACK, ACTION_BACK_NAME),
        ]
    )
    def intent_mapping(self, request):
        return request.param

    def test_predict_mapped_action(self, priority, domain_with_mapping, intent_mapping):
        policy = self.create_policy(None, priority)
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered(intent_mapping[0], 1),
        ]

        assert (
            self._get_next_action(policy, events, domain_with_mapping)
            == intent_mapping[1]
        )

    def test_restart_if_paused(self, priority, domain_with_mapping):
        policy = self.create_policy(None, priority)
        events = [ConversationPaused(), user_uttered(USER_INTENT_RESTART, 1)]

        assert (
            self._get_next_action(policy, events, domain_with_mapping)
            == ACTION_RESTART_NAME
        )

    def test_predict_action_listen(self, priority, domain_with_mapping, intent_mapping):
        policy = self.create_policy(None, priority)
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered(intent_mapping[0], 1),
            ActionExecuted(intent_mapping[1], policy="policy_0_MappingPolicy"),
        ]
        tracker = get_tracker(events)
        scores = policy.predict_action_probabilities(tracker, domain_with_mapping)
        index = scores.index(max(scores))
        action_planned = domain_with_mapping.action_names[index]
        assert action_planned == ACTION_LISTEN_NAME
        assert scores != [0] * domain_with_mapping.num_actions

    def test_do_not_follow_other_policy(
        self, priority, domain_with_mapping, intent_mapping
    ):
        policy = self.create_policy(None, priority)
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered(intent_mapping[0], 1),
            ActionExecuted(intent_mapping[1], policy="other_policy"),
        ]
        tracker = get_tracker(events)
        scores = policy.predict_action_probabilities(tracker, domain_with_mapping)
        assert scores == [0] * domain_with_mapping.num_actions


class TestFallbackPolicy(PolicyTestCollection):
    def create_policy(self, featurizer, priority):
        p = FallbackPolicy(priority=priority)
        return p

    def test_featurizer(self, trained_policy, tmpdir):
        assert trained_policy.featurizer is None
        trained_policy.persist(tmpdir.strpath)
        loaded = trained_policy.__class__.load(tmpdir.strpath)
        assert loaded.featurizer is None

    @pytest.mark.parametrize(
        "top_confidence, all_confidences, last_action_name, should_nlu_fallback",
        [
            (0.1, [0.1], "some_action", False),
            (0.1, [0.1], "action_listen", True),
            (0.9, [0.9, 0.1], "some_action", False),
            (0.9, [0.9, 0.1], "action_listen", False),
            (0.4, [0.4, 0.35], "some_action", False),
            (0.4, [0.4, 0.35], "action_listen", True),
            (0.9, [0.9, 0.85], "action_listen", True),
        ],
    )
    def test_should_nlu_fallback(
        self,
        trained_policy,
        top_confidence,
        all_confidences,
        last_action_name,
        should_nlu_fallback,
    ):
        nlu_data = {
            "intent": {"confidence": top_confidence},
            "intent_ranking": [
                {"confidence": confidence} for confidence in all_confidences
            ],
        }
        assert (
            trained_policy.should_nlu_fallback(nlu_data, last_action_name)
            is should_nlu_fallback
        )


class TestTwoStageFallbackPolicy(TestFallbackPolicy):
    def create_policy(self, featurizer, priority):
        p = TwoStageFallbackPolicy(
            priority=priority, deny_suggestion_intent_name="deny"
        )
        return p

    @pytest.fixture(scope="class")
    def default_domain(self):
        content = """
        intents:
          - greet
          - bye
          - affirm
          - deny
        """
        return Domain.from_yaml(content)

    @staticmethod
    async def _get_tracker_after_reverts(events, channel, nlg, domain):
        tracker = get_tracker(events)
        action = ActionRevertFallbackEvents()
        events += await action.run(channel, nlg, tracker, domain)

        return get_tracker(events)

    def test_ask_affirmation(self, trained_policy, default_domain):
        events = [ActionExecuted(ACTION_LISTEN_NAME), user_uttered("Hi", 0.2)]

        next_action = self._get_next_action(trained_policy, events, default_domain)

        assert next_action == ACTION_DEFAULT_ASK_AFFIRMATION_NAME

    async def test_affirmation(self, default_channel, default_nlg, default_domain):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 1),
            ActionExecuted("utter_hello"),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 1),
        ]

        tracker = await self._get_tracker_after_reverts(
            events, default_channel, default_nlg, default_domain
        )

        assert "greet" == tracker.latest_message.parse_data["intent"]["name"]
        assert tracker.export_stories() == (
            "## sender\n* greet\n    - utter_hello\n* greet\n"
        )

    def test_ask_rephrase(self, trained_policy, default_domain):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("deny", 1),
        ]

        next_action = self._get_next_action(trained_policy, events, default_domain)

        assert next_action == ACTION_DEFAULT_ASK_REPHRASE_NAME

    async def test_successful_rephrasing(
        self, default_channel, default_nlg, default_domain
    ):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("deny", 1),
            ActionExecuted(ACTION_DEFAULT_ASK_REPHRASE_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("bye", 1),
        ]

        tracker = await self._get_tracker_after_reverts(
            events, default_channel, default_nlg, default_domain
        )

        assert "bye" == tracker.latest_message.parse_data["intent"]["name"]
        assert tracker.export_stories() == "## sender\n* bye\n"

    def test_affirm_rephrased_intent(self, trained_policy, default_domain):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("deny", 1),
            ActionExecuted(ACTION_DEFAULT_ASK_REPHRASE_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
        ]

        next_action = self._get_next_action(trained_policy, events, default_domain)

        assert next_action == ACTION_DEFAULT_ASK_AFFIRMATION_NAME

    async def test_affirmed_rephrasing(
        self, default_channel, default_nlg, default_domain
    ):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("deny", 1),
            ActionExecuted(ACTION_DEFAULT_ASK_REPHRASE_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("bye", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("bye", 1),
        ]

        tracker = await self._get_tracker_after_reverts(
            events, default_channel, default_nlg, default_domain
        )

        assert "bye" == tracker.latest_message.parse_data["intent"]["name"]
        assert tracker.export_stories() == "## sender\n* bye\n"

    def test_denied_rephrasing_affirmation(self, trained_policy, default_domain):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("deny", 1),
            ActionExecuted(ACTION_DEFAULT_ASK_REPHRASE_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("bye", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("deny", 1),
        ]

        next_action = self._get_next_action(trained_policy, events, default_domain)

        assert next_action == ACTION_DEFAULT_FALLBACK_NAME

    async def test_rephrasing_instead_affirmation(
        self, default_channel, default_nlg, default_domain
    ):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 1),
            ActionExecuted("utter_hello"),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("bye", 1),
        ]

        tracker = await self._get_tracker_after_reverts(
            events, default_channel, default_nlg, default_domain
        )

        assert "bye" == tracker.latest_message.parse_data["intent"]["name"]
        assert tracker.export_stories() == (
            "## sender\n* greet\n    - utter_hello\n* bye\n"
        )

    def test_unknown_instead_affirmation(self, trained_policy, default_domain):
        events = [
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
            ActionExecuted(ACTION_DEFAULT_ASK_AFFIRMATION_NAME),
            ActionExecuted(ACTION_LISTEN_NAME),
            user_uttered("greet", 0.2),
        ]

        next_action = self._get_next_action(trained_policy, events, default_domain)

        assert next_action == ACTION_DEFAULT_FALLBACK_NAME

    def test_listen_after_hand_off(self, trained_policy, default_domain):
        events = [ActionExecuted(ACTION_DEFAULT_FALLBACK_NAME)]

        next_action = self._get_next_action(trained_policy, events, default_domain)

        assert next_action == ACTION_LISTEN_NAME



import numpy as np
import pytest

from rasa.nlu.tokenizers.tokenizer import Tokenizer
from rasa.nlu.training_data import TrainingData
from rasa.nlu.tokenizers.convert_tokenizer import ConveRTTokenizer
from rasa.nlu.constants import (
    TEXT,
    DENSE_FEATURE_NAMES,
    TOKENS_NAMES,
    RESPONSE,
    INTENT,
)
from rasa.nlu.training_data import Message
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.featurizers.dense_featurizer.convert_featurizer import ConveRTFeaturizer


def test_convert_featurizer_process():
    featurizer = ConveRTFeaturizer.create({}, RasaNLUModelConfig())

    sentence = "Hey how are you today ?"
    message = Message(sentence)
    tokens = ConveRTTokenizer().tokenize(message, attribute=TEXT)
    tokens = Tokenizer.add_cls_token(tokens, attribute=TEXT)
    message.set(TOKENS_NAMES[TEXT], tokens)

    featurizer.process(message)

    expected = np.array([2.2636216, -0.26475656, -1.1358104, -0.49751878, -1.3946456])
    expected_cls = np.array(
        [1.0251294, -0.04053932, -0.7018805, -0.82054937, -0.75054353]
    )

    vecs = message.get(DENSE_FEATURE_NAMES[TEXT])

    assert len(tokens) == len(vecs)
    assert np.allclose(vecs[0][:5], expected, atol=1e-5)
    assert np.allclose(vecs[-1][:5], expected_cls, atol=1e-5)


def test_convert_featurizer_train():
    featurizer = ConveRTFeaturizer.create({}, RasaNLUModelConfig())

    sentence = "Hey how are you today ?"
    message = Message(sentence)
    message.set(RESPONSE, sentence)
    tokens = ConveRTTokenizer().tokenize(message, attribute=TEXT)
    tokens = Tokenizer.add_cls_token(tokens, attribute=TEXT)
    message.set(TOKENS_NAMES[TEXT], tokens)
    message.set(TOKENS_NAMES[RESPONSE], tokens)

    featurizer.train(TrainingData([message]), RasaNLUModelConfig())

    expected = np.array([2.2636216, -0.26475656, -1.1358104, -0.49751878, -1.3946456])
    expected_cls = np.array(
        [1.0251294, -0.04053932, -0.7018805, -0.82054937, -0.75054353]
    )

    vecs = message.get(DENSE_FEATURE_NAMES[TEXT])

    assert len(tokens) == len(vecs)
    assert np.allclose(vecs[0][:5], expected, atol=1e-5)
    assert np.allclose(vecs[-1][:5], expected_cls, atol=1e-5)

    vecs = message.get(DENSE_FEATURE_NAMES[RESPONSE])

    assert len(tokens) == len(vecs)
    assert np.allclose(vecs[0][:5], expected, atol=1e-5)
    assert np.allclose(vecs[-1][:5], expected_cls, atol=1e-5)

    vecs = message.get(DENSE_FEATURE_NAMES[INTENT])

    assert vecs is None


@pytest.mark.parametrize(
    "sentence, expected_text",
    [
        ("hello", "hello"),
        ("you're", "you re"),
        ("r. n. b.", "r n b"),
        ("rock & roll", "rock & roll"),
        ("ńöñàśçií", "ńöñàśçií"),
    ],
)
def test_convert_featurizer_tokens_to_text(sentence, expected_text):
    tokens = ConveRTTokenizer().tokenize(Message(sentence), attribute=TEXT)

    actual_text = ConveRTFeaturizer._tokens_to_text([tokens])[0]

    assert expected_text == actual_text



from rasa.core import run

CREDENTIALS_FILE = "examples/moodbot/credentials.yml"


def test_create_http_input_channels():
    channels = run.create_http_input_channels(None, CREDENTIALS_FILE)
    assert len(channels) == 7

    # ensure correct order
    assert {c.name() for c in channels} == {
        "twilio",
        "slack",
        "telegram",
        "mattermost",
        "facebook",
        "webexteams",
        "rocketchat",
    }


def test_create_single_input_channels():
    channels = run.create_http_input_channels("facebook", CREDENTIALS_FILE)
    assert len(channels) == 1
    assert channels[0].name() == "facebook"


def test_create_single_input_channels_by_class():
    channels = run.create_http_input_channels(
        "rasa.core.channels.channel.RestInput", CREDENTIALS_FILE
    )
    assert len(channels) == 1
    assert channels[0].name() == "rest"


def test_create_single_input_channels_by_class_wo_credentials():
    channels = run.create_http_input_channels(
        "rasa.core.channels.channel.RestInput", credentials_file=None
    )

    assert len(channels) == 1
    assert channels[0].name() == "rest"



from rasa.nlu.components import Component
from rasa.nlu import utils
from rasa.nlu.model import Metadata
import os
import spacy
import pandas as pd
import typing
from typing import Any, Optional, Text, Dict
import numpy as np
import regex as re
from rasa.nlu.training_data import Message, TrainingData


class Extracteur_oncf(Component):
	"""A custom sentiment analysis component"""
	name = "DATA_EXTRACTION"
	provides = ["entities"]
	requires = ["tokens"]
	defaults = {}
	language_list = ["en"]
	print('initialised the class')

	def _init_(self, component_config=None):
		super(Extracteur_omran, self)._init_(component_config)

	def train(self, training_data, cfg, **kwargs):
		"""Load the sentiment polarity labels from the text
		   file, retrieve training tokens and after formatting
		   data train the classifier."""



	def convert_to_rasa_gare(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "surface",
				"extractor": "extractor"}

		return entity
    
	def convert_to_rasa_prix(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""
		
		entity = {"value": value,
				  
				"entity": "price",
				"extractor": "extractor"}

		return entity

	def convert_to_rasa_localite(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "localite",
				"extractor": "extractor"}

		return entity
	def convert_to_rasa_region(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "region",
				"extractor": "extractor"}

		return entity

	def convert_to_rasa_type(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "type",
				"extractor": "extractor"}

		return entity
		


	def process(self, message:Message , **kwargs):
		"""Retrieve the tokens of the new message, pass it to the classifier
			and append prediction results to the message class."""
		print(message.text)	
		if not self :
			entities = []
		else:
			
#detection cities	
			tokens = [t.text for t in message.get("tokens")]
			print('***********tokens*****')
			print(tokens)
			datas = pd.read_csv('csv_files/villes.csv',sep=';',encoding="utf_8")
			Tville= np.array(datas['ville'])
			Oville= np.array(datas['value'])
			ent_val = {}
			entity_conv_city = []
			entity_city = ""
			for i in range(len(Tville)):
				ent_val[Tville[i]] = Oville[i]
			for word in tokens:
				for key in ent_val:
					if key == word:
						entity_city = ent_val[key]
						print(" extracted Token ++++++++++ "+ word)
						
			entity_conv_city = self.convert_to_rasa_localite(entity_city)

#detection gare	
			tokens = [t.text for t in message.get("tokens")]
			print('***********tokens*****')
			print(tokens)
			datas = pd.read_csv('csv_files/regions.csv',sep=';',encoding="utf_8")
			Tville= np.array(datas['ville'])
			Oville= np.array(datas['value'])
			ent_val = {}
			entity_conv_region = []
			entity_region = ""
			for i in range(len(Tville)):
				ent_val[Tville[i]] = Oville[i]
			for word in tokens:
				for key in ent_val:
					if key == word:
						entity_region = ent_val[key]
						print(" extracted Token ++++++++++ "+ word)
						
			entity_conv_region = self.convert_to_rasa_region(entity_region)

#detection de surface !!!!
			tokens = [t.text for t in message.get("tokens")]
			ent_val = {}
			entity_conv_surf = []
			entity_surf = ""
            # liste des expressions
			exp = ['متر' ,'م','m²','m','mètre','mètres','metre','metres','ما']
			for word in tokens:
				if word.isdigit():
					if tokens[tokens.index(word)+1] in exp:
						print("l9ina surface")
						entity_surf = word
                    
			entity_conv_surf = self.convert_to_rasa_surf(entity_surf)
#detection de budget

			tokens = [t.text for t in message.get("tokens")]
			#print("flooooos")
			Prix = ['درهم','ده','دهس','dh','dhs','dirham','dirhams']
			million=['مليون','million']
			centime = ['سنتيم','centime']
			entity_conv_prix = []
			entity_prix = ""
			for word in tokens:
				print(" ALL Tokeens ++++++++++ "+ word)
				if word.isdigit():
					print("klmaaaaaaaaaaaaaaaa " + str(tokens.index(word)+1))
					if tokens[tokens.index(word)+1] in million:
						print("milooooooooooooooooooooooon " + word)
						entity_prix = str(int(word)*1000000)
					if tokens[tokens.index(word)+1] in centime:
						entity_prix = str(int(word)/100)	
					if tokens[tokens.index(word)+1] in Prix:
						entity_prix = word
						print("prices++++++++++++++++++ "+entity_prix)

			entity_conv_prix = self.convert_to_rasa_prix(entity_prix)

#detection type produit
			tokens = [t.text for t in message.get("tokens")]
			data_prod = pd.read_csv('csv_files/type_produit.csv',sep=';',encoding="utf_8")
			type_prod= np.array(data_prod['name'])
			prod= np.array(data_prod['value'])
			ent_val2 = {}
			entity_conv2 = []
			entity2 = ""
			for i in range(len(type_prod)):
				ent_val2[type_prod[i]] = prod[i]
			for w in tokens:
				for key2 in ent_val2:
					if key2 == w:
						entity2 = ent_val2[key2]

			entity_conv2 = self.convert_to_rasa_type(entity2)			

			message.set("entity_city", [entity_conv_city], add_to_output=True)
			message.set("entity_region", [entity_conv_region], add_to_output=True)
			message.set("entity_surface", [entity_conv_surf], add_to_output=True)
			message.set("entity_type", [entity_conv2], add_to_output=True)
			message.set("entity_prix", [entity_conv_prix], add_to_output=True)


import argparse
import asyncio
import logging
from typing import Text

import os
import rasa.utils.io
import rasa.train
from examples.restaurantbot.policy import RestaurantPolicy
from rasa.core.agent import Agent
from rasa.core.policies.memoization import MemoizationPolicy
from rasa.core.policies.mapping_policy import MappingPolicy

logger = logging.getLogger(__name__)


async def parse(text: Text, model_path: Text):
    agent = Agent.load(model_path)

    response = await agent.handle_text(text)

    logger.info(f"Text: '{text}'")
    logger.info("Response:")
    logger.info(response)

    return response


async def train_core(
    domain_file: Text = "domain.yml",
    model_directory: Text = "models",
    model_name: Text = "current",
    training_data_file: Text = "data/stories.md",
):
    agent = Agent(
        domain_file,
        policies=[
            MemoizationPolicy(max_history=3),
            MappingPolicy(),
            RestaurantPolicy(batch_size=100, epochs=100, validation_split=0.2),
        ],
    )

    training_data = await agent.load_data(training_data_file, augmentation_factor=10)
    agent.train(training_data)

    # Attention: agent.persist stores the model and all meta data into a folder.
    # The folder itself is not zipped.
    model_path = os.path.join(model_directory, model_name, "core")
    agent.persist(model_path)

    logger.info(f"Model trained. Stored in '{model_path}'.")

    return model_path


def train_nlu(
    config_file="config.yml",
    model_directory: Text = "models",
    model_name: Text = "current",
    training_data_file="data/nlu.md",
):
    from rasa.nlu.training_data import load_data
    from rasa.nlu import config
    from rasa.nlu.model import Trainer

    training_data = load_data(training_data_file)
    trainer = Trainer(config.load(config_file))
    trainer.train(training_data)

    # Attention: trainer.persist stores the model and all meta data into a folder.
    # The folder itself is not zipped.
    model_path = os.path.join(model_directory, model_name)
    model_directory = trainer.persist(model_path, fixed_model_name="nlu")

    logger.info(f"Model trained. Stored in '{model_directory}'.")

    return model_directory


if __name__ == "__main__":
    rasa.utils.io.configure_colored_logging(loglevel="INFO")

    parser = argparse.ArgumentParser(description="Restaurant Bot")

    subparser = parser.add_subparsers(dest="subparser_name")
    train_parser = subparser.add_parser("train", help="train a core or nlu model")
    parse_parser = subparser.add_parser("predict", help="predict next action")

    parse_parser.add_argument(
        "--model",
        default="models/current",
        help="Path to the model directory which contains "
        "sub-folders for core and nlu models.",
    )
    parse_parser.add_argument("--text", default="hello", help="Text to parse.")

    train_parser.add_argument(
        "model",
        choices=["nlu", "core"],
        help="Do you want to train a NLU or Core model?",
    )
    args = parser.parse_args()

    loop = asyncio.get_event_loop()

    # decide what to do based on first parameter of the script
    if args.subparser_name == "train":
        if args.model == "nlu":
            train_nlu()
        elif args.model == "core":
            loop.run_until_complete(train_core())
    elif args.subparser_name == "predict":
        loop.run_until_complete(parse(args.text, args.model))



def test_dummy_request():
    from rasa.nlu.emulators.no_emulator import NoEmulator

    em = NoEmulator()
    norm = em.normalise_request_json({"text": ["arb text"]})
    assert norm == {"text": "arb text", "time": None}

    norm = em.normalise_request_json({"text": ["arb text"], "time": "1499279161658"})
    assert norm == {"text": "arb text", "time": "1499279161658"}


def test_dummy_response():
    from rasa.nlu.emulators.no_emulator import NoEmulator

    em = NoEmulator()
    data = {"intent": "greet", "text": "hi", "entities": {}, "confidence": 1.0}
    assert em.normalise_response_json(data) == data


def test_emulators_can_handle_missing_data():
    from rasa.nlu.emulators.luis import LUISEmulator

    em = LUISEmulator()
    norm = em.normalise_response_json(
        {"text": "this data doesn't contain an intent result"}
    )
    assert norm["topScoringIntent"] is None
    assert norm["intents"] == []



TEXT = "text"

RESPONSE_KEY_ATTRIBUTE = "response_key"

INTENT = "intent"

RESPONSE = "response"

ENTITIES = "entities"
BILOU_ENTITIES = "bilou_entities"
NO_ENTITY_TAG = "O"

EXTRACTOR = "extractor"

PRETRAINED_EXTRACTORS = {"DucklingHTTPExtractor", "SpacyEntityExtractor"}

CLS_TOKEN = "__CLS__"

MESSAGE_ATTRIBUTES = [TEXT, INTENT, RESPONSE]

TOKENS_NAMES = {TEXT: "tokens", INTENT: "intent_tokens", RESPONSE: "response_tokens"}

SPARSE_FEATURE_NAMES = {
    TEXT: "text_sparse_features",
    INTENT: "intent_sparse_features",
    RESPONSE: "response_sparse_features",
}

DENSE_FEATURE_NAMES = {
    TEXT: "text_dense_features",
    INTENT: "intent_dense_features",
    RESPONSE: "response_dense_features",
}

LANGUAGE_MODEL_DOCS = {
    TEXT: "text_language_model_doc",
    RESPONSE: "response_language_model_doc",
}

TOKEN_IDS = "token_ids"
TOKENS = "tokens"
SEQUENCE_FEATURES = "sequence_features"
SENTENCE_FEATURES = "sentence_features"

SPACY_DOCS = {TEXT: "text_spacy_doc", RESPONSE: "response_spacy_doc"}


DENSE_FEATURIZABLE_ATTRIBUTES = [TEXT, RESPONSE]

RESPONSE_SELECTOR_PROPERTY_NAME = "response_selector"
DEFAULT_OPEN_UTTERANCE_TYPE = "default"
OPEN_UTTERANCE_PREDICTION_KEY = "response"
OPEN_UTTERANCE_RANKING_KEY = "ranking"
RESPONSE_IDENTIFIER_DELIMITER = "/"



# This files contains your custom actions which can be used to run
# custom Python code.
#
# See this guide on how to implement these action:
# https://rasa.com/docs/rasa/core/actions/#custom-actions/


# This is a simple example for an assistant that schedules reminders and
# reacts to external events.

from typing import Any, Text, Dict, List
import datetime

from rasa_sdk import Action, Tracker
from rasa_sdk.events import ReminderScheduled, ReminderCancelled
from rasa_sdk.executor import CollectingDispatcher


class ActionSetReminder(Action):
    """Schedules a reminder, supplied with the last message's entities."""

    def name(self) -> Text:
        return "action_set_reminder"

    async def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:

        dispatcher.utter_message("I will remind you in 5 seconds.")

        date = datetime.datetime.now() + datetime.timedelta(seconds=5)
        entities = tracker.latest_message.get("entities")

        reminder = ReminderScheduled(
            "EXTERNAL_reminder",
            trigger_date_time=date,
            entities=entities,
            name="my_reminder",
            kill_on_user_message=False,
        )

        return [reminder]


class ActionReactToReminder(Action):
    """Reminds the user to call someone."""

    def name(self) -> Text:
        return "action_react_to_reminder"

    async def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:

        name = next(tracker.get_latest_entity_values("name"), "someone")
        dispatcher.utter_message(f"Remember to call {name}!")

        return []


class ActionTellID(Action):
    """Informs the user about the conversation ID."""

    def name(self) -> Text:
        return "action_tell_id"

    async def run(
        self, dispatcher, tracker: Tracker, domain: Dict[Text, Any]
    ) -> List[Dict[Text, Any]]:

        conversation_id = tracker.sender_id

        dispatcher.utter_message(
            f"The ID of this conversation is: " f"{conversation_id}."
        )

        dispatcher.utter_message(
            f"Trigger an intent with "
            f'curl -H "Content-Type: application/json" '
            f'-X POST -d \'{{"name": "EXTERNAL_dry_plant", '
            f'"entities": {{"plant": "Orchid"}}}}\' '
            f"http://localhost:5005/conversations/{conversation_id}/"
            f"trigger_intent"
        )

        return []


class ActionWarnDry(Action):
    """Informs the user that a plant needs water."""

    def name(self) -> Text:
        return "action_warn_dry"

    async def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:

        plant = next(tracker.get_latest_entity_values("plant"), "someone")
        dispatcher.utter_message(f"Your {plant} needs some water!")

        return []


class ForgetReminders(Action):
    """Cancels all reminders."""

    def name(self) -> Text:
        return "action_forget_reminders"

    async def run(
        self, dispatcher, tracker: Tracker, domain: Dict[Text, Any]
    ) -> List[Dict[Text, Any]]:

        dispatcher.utter_message(f"Okay, I'll cancel all your reminders.")

        # Cancel all reminders
        return [ReminderCancelled()]



from builtins import vars
from rasa.nlu.components import Component
from rasa.nlu import utils
from rasa.nlu.model import Metadata
import os
import pandas as pd
import typing

from textblob import TextBlob
import json
from typing import Any, Optional, Text, Dict
import numpy as np
import regex as re
from time import strftime
from datetime import datetime as dt
from typing import Any, Optional, Text, Dict
import numpy as np
import dateutil.parser as dparser
import datefinder

from dateutil.parser import parse
import datetime
import dateutil.parser as dparser
from pyarabic.number import text2number
import datefinder
from dateutil.parser import parse
from datetime import timedelta
from dateutil.relativedelta import relativedelta


from rasa.nlu.training_data import Message, TrainingData


class Extractor_ecommerce_ar(Component):
	"""A custom Ecom analysis component"""
	name = "Ecommerce_EXTRACTION"
	provides = ["entities"]
	requires = ["tokens"]
	defaults = {}
	language_list = ["en","fr"]
	print('initialised the class')

	def _init_(self, component_config=None):
		super(Extracteur_oncf, self)._init_(component_config)

	def train(self, training_data, cfg, **kwargs):
		"""Load the sentiment polarity labels from the text
		   file, retrieve training tokens and after formatting
		   data train the classifier."""

	
	def convert_to_size(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "SIZE",
				  "extractor": "extractor"}

		return entity
	
	def convert_to_amount_money(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "AMOUNT_MONEY",
				  "extractor": "extractor"}

		return entity
	
	def convert_to_currency(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "CURRENCY",
				  "extractor": "extractor"}

		return entity
	
	def convert_to_home_deco(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "HOME_DECO",
				  "extractor": "extractor"}

		return entity
	
	def convert_to_metal(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "METAL",
				  "extractor": "extractor"}

		return entity

	def convert_to_amount(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "AMNT_MONEY",
				  "extractor": "extractor"}

		return entity

	def convert_to_fabric(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "FABRIC",
				  "extractor": "extractor"}

		return entity

	def convert_to_clothes(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "CLOTHES",
				  "extractor": "extractor"}

		return entity
	def convert_to_color(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "COLOR",
				  "extractor": "extractor"}

		return entity	
	
	

	def convert_to_number(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "NUMBER",
				  "extractor": "extractor"}

		return entity


	def convert_to_electro(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "ELECTRO",
				  "extractor": "extractor"}

		return entity
	
	def convert_to_access(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "ACCESS",
				  "extractor": "extractor"}

		return entity
	def convert_to_cosm(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "COSMETICS",
				  "extractor": "extractor"}

		return entity
	
	def convert_to_category(self, dict):
		"""Convert model output into the Rasa NLU compatible output format.""" 
		entities = []
		for i in range(0,len(dict)):
			
			entity = {	"value": dict[i]["value"],
				  
				  		"category": dict[i]["type"],

				  		"extractor": "extractor_categories_products"}

			entities.append(entity)

		return entities
	
	def convert_to_variation(self, dict):
		"""Convert model output into the Rasa NLU compatible output format.""" 
		entities = []
		for i in range(0,len(dict)):
			
			entity = {	"value": dict[i]["value"],
				  
				  		"category": dict[i]["type"],

				  		"extractor": "extractor_categories_products"}

			entities.append(entity)

		return entities


	def process(self, message:Message , **kwargs):
		"""Retrieve the tokens of the new message, pass it to the classifier
			and append prediction results to the message class."""
		print(message.text)	
		if not self :
			entities = []
		else:
			tokens3 = message.text.split()
			tokens = [t.text for t in message.get("tokens")]
			print('***********tokens*****')
			print(tokens)
			
			
			
			entities_clothes = []
			entities_size = []
			entities_color = []
			entities_electro = []
			entities_metal = []
			entities_fabric = []
			entities_home_deco = []
			entities_access = []

			
			Electro_keys = ["COMPUTER","TV","PHONE","BEAUTYELECTRO","PHOTO","AUDIO","GAME","KITCHEN","TAB"]
			Electro_vals = '{ "computer":{"ar": ["بس ","حاسوب","كومبتر"] ,"fr":["pc","computer","ordinateur"],"ang":["computer"," pc "]} , "tv" : {"ar": ["تلفاز","تلفزيون"] ,"fr":["télé","tv","television"],"ang":[" tv "]},"phone": {"ar": ["تلفون","هاتف","سمارتفون"] ,"fr":["telephone","smartphone","portable","tel","iphone"],"ang":["phone","tablet"]},"beautyelectro" : {"ar": ["سشوار","ليسور","بروس","ابيلاتور","بروس"] ,"fr":["lisseur","sechoir","epilateur","tondeuse","seche.cheveu"],"ang":["brush","epilator","hair.straightener"]},"photo" : {"ar": ["كاميرا"] ,"fr":["appareil.photo"],"ang":["camera"]},"audio" :{"ar": ["سماعة","سماعات","اكوتور"] ,"fr":["casque","ecouteurs"],"ang":["headphone","earphone"]},"game":{"ar": ["بلاي"] ,"fr":["play"],"ang":["playstation"]},"kitchen" :{"ar": ["ميكسور"] ,"fr":["batteur"],"ang":["mixer"]},"tab":{"ar": ["تابليت"] ,"fr":["tablette"],"ang":["tablet"]}}' 
			msg = ''

			for t in tokens :
				if t == "__CLS__" :
					print(" ")
				else :
					msg = msg+" "+t

			list_electro = json.loads(Electro_vals)

			print(Electro_keys[0].lower())
		
			electro_fr = []
			electro_ar = []
			electro_ang = []
			group_types = []
			keys_types = []
			categories = []
			for key in Electro_keys :
				
				
					
				for i in range(len(list_electro[key.lower()]["ar"])):

					type_ar = re.search(list_electro[key.lower()]["ar"][i],msg)
					if type_ar != None :
						group_types.append(type_ar.group(0))
						electro_ar.append(key)
						keys_types.append(key)
						categories.append("ELECTRO")
						break
				
			
			if electro_ar != []:
				for i in range(len(electro_ar)):

					entities_electro.append(electro_ar[i])
			
			
			entities_electro = list(dict.fromkeys(entities_electro))
			
			
			print(entities_electro)
			while("" in entities_electro) :
				entities_electro.remove("")
			

			entity_electro = self.convert_to_electro(entities_electro)
			

			dict_entities_category = []
			






			Watch_acc_keys= ["WATCH","NECKLASS","BRACELET","RING","EARRINGS","SCARF","BELT","GLASSES","GLOVE","TIE","HAT"]
			
						
			Watch_vals = '{ "watch":{"ar": ["ساعه"] ,"fr":["montre"],"ang":["watch"]} , "necklass" : {"ar": ["قلادة","كولير"] ,"fr":["collier"],"ang":["necklace"]},"bracelet": {"ar": ["براسليت","كورميت"] ,"fr":["gourmette","bracelet"],"ang":["bracelet"]},"ring" : {"ar": ["خاتم","باج"] ,"fr":["bague"],"ang":["ring"]},"earrings" : {"ar": ["حلق","بوكل"] ,"fr":["boucle.oreille"],"ang":["earring"]},"scarf" :{"ar": ["وشاح","فولار","اشارب"] ,"fr":["foulard","echarpe"],"ang":["scarf"]},"belt":{"ar": ["حزام","سمطه"] ,"fr":["ceinture"],"ang":["belt"]},"glasses" :{"ar": ["ندادر","نظارات"] ,"fr":["lunette"],"ang":["glasses","sunglasses"]},"glove":{"ar": ["قفاز"] ,"fr":["gant"],"ang":["glove"]},"tie":{"ar": ["ربطة.عنق"] ,"fr":["cravate"],"ang":["tie"]},"hat":{"ar": ["كاسكيت","شابو"] ,"fr":["chapeau"],"ang":["hat"]} }' 

			list_access = json.loads(Watch_vals)
			msg = ''

			for t in tokens :
				if t == "__CLS__" :
					print(" ")
				else :
					msg = msg+" "+t+" "
			access_fr = []
			access_ar = []
			access_ang = []
			
			for key in Watch_acc_keys :
				
				
					
				for i in range(len(list_access[key.lower()]["ar"])):

					type_ar = re.search(list_access[key.lower()]["ar"][i],msg)
					if type_ar != None :
						group_types.append(type_ar.group(0))
						keys_types.append(key)
						access_ar.append(key)
						categories.append("WATCH_ACC")
						break
				
			
			
			

			print("access_fr")
			print(access_fr)
			print("access_ar")
			print(access_ar)
			
			print(entities_access)
			if access_ar != []:
				for i in range(len(access_ar)):
					entities_access.append(access_ar[i])
			
			print(entities_access)
			
			entities_access = list(dict.fromkeys(entities_access))
			
			
			
			while("" in entities_access) :
				entities_access.remove("")

			


			entity_access = self.convert_to_access(entities_access)
			



			Home_deco_keys = ["CANDLE","SHEET","PILLOW","LAMP","FURNITURE","MIRROR"]
			


			Home_deco_vals = '{ "candle":{"ar": ["شمعه"] ,"fr":["bougie"],"ang":["candle"]} , "sheet" : {"ar": ["لحاف"] ,"fr":["couette","drap"],"ang":["sheet","blanket"]},"pillow": {"ar": ["وسادة","مخدة"] ,"fr":["taie"],"ang":["pillow"]},"lamp" : {"ar": ["مصباح","لامبه"] ,"fr":["lampe"],"ang":["lamp"]},"furniture" : {"ar": ["موبل"] ,"fr":["meuble"],"ang":["furniture"]},"mirror" : {"ar": ["مرايا","ميروار"] ,"fr":["mirroir"],"ang":["mirror"]}}' 

			list_home_deco = json.loads(Home_deco_vals)


			home_fr = []
			home_ar = []
			home_ang = []
			
			for key in Home_deco_keys :
				
				
					
				for i in range(len(list_home_deco[key.lower()]["ar"])):

					type_ar_home = re.search(list_home_deco[key.lower()]["ar"][i],msg)
					if type_ar_home != None :
						home_ar.append(key)
						keys_types.append(key)
						group_types.append(type_ar_home.group(0))
						categories.append("HOME_DECO")
						break
				
			
			
			

			
			print("home_ar")
			print(home_ar)
			print("home_ang")
			print(home_ang)
			
			if home_ar != "":
				for i in range(len(home_ar)):
					
					entities_home_deco.append(home_ar[i])
			
					
					entities_home_deco.append(home_ang[i])
			print(entities_home_deco)
			
			entities_home_deco = list(dict.fromkeys(entities_home_deco))
			
			
			print(entities_home_deco)
			while("" in entities_home_deco) :
				entities_home_deco.remove("")

			entity_home_deco = self.convert_to_home_deco(entities_home_deco)
			


			keys_vars = []

			Metal_keys = ["GOLD","IRON","INOX","SILVER","ALUMINIUM","BRONZE","PLAQUEOR"]
			


			Metal_vals = '{ "gold":{"ar": ["ذهب"] ,"fr":[" or "],"ang":["gold"]} , "iron" : {"ar": ["حديد"] ,"fr":["fer"],"ang":["iron"]},"inox": {"ar": ["انوكس"] ,"fr":["inox","inoxidable"],"ang":["stainless.steel"]},"silver" : {"ar": ["فضة","ارجونت"] ,"fr":["argent"],"ang":["silver"]},"aluminium" : {"ar": ["الومنيوم"] ,"fr":["aluminium"],"ang":["aluminium"]},"bronze" : {"ar": ["برونز"] ,"fr":["bronze"],"ang":["bronze"]},"plaqueor" : {"ar": ["بلاكيور"] ,"fr":["plaque.or"],"ang":["gold.plate"]}}' 

			list_metal = json.loads(Metal_vals)
			variations =[]

			group_vars = []
			metal_fr = []
			metal_ar = []
			metal_ang = []
			for key in Metal_keys :
				
				
					
				for i in range(len(list_metal[key.lower()]["ar"])):

					type_ar = re.search(list_metal[key.lower()]["ar"][i],msg)
					if type_ar != None :
						metal_ar.append(key)
						keys_vars.append(key)
						group_vars.append(type_ar.group(0))
						variations.append("METAL")
						break
				
			
			
			if metal_ar != "":
				for i in range(len(metal_ar)):
					
					entities_metal.append(metal_ar[i])
			
			
			entities_metal = list(dict.fromkeys(entities_metal))
			
			while("" in entities_metal) :
				entities_metal.remove("")

			entity_metal = self.convert_to_metal(entities_metal)
			





			Clothes_keys = ["TOP","DRESS","BLAZER","PANTS","SKIRT","SHOES","PYJAMA","BAG","CAFTAN"]
			


			Clothes_vals = '{ "top":{"ar": ["تشيرت", "بلوزة","شوميز"] ,"fr":["tshirt","chemise","haut","pull","blouse","capuche"],"ang":["shirt","blouse","sweater"]} , "dress" : {"ar": ["كسوة","روب"] ,"fr":["robe"],"ang":["dress"]},"blazer": {"ar": ["معطف","فست","جاكيت"] ,"fr":["veste","manteau","doudoune"],"ang":["coat"]},"pants" : {"ar": ["سروال","جين"] ,"fr":["pantalon","short","legging","jogging"],"ang":["jeans"]},"skirt" : {"ar": ["تنورة","صاية","جوب"] ,"fr":["jupe"],"ang":["skirt"]},"shoes" : {"ar": ["سباط","سنداله","طالون","سبرديلا","ساندال","اسبا","اسكاربين"] ,"fr":["bottes","chaussure","basket","espas","talons","sandale"],"ang":["shoes","sneakers"]},"pyjama" : {"ar":["بيجاما"] ,"fr":["pyjama"],"ang":["pyjama"]},"bag" : {"ar": ["ساك"] ,"fr":["sac","cartable","pochette","sacoche","sac.dos"],"ang":["purse","bag"]}, "caftan" : {"ar": ["قفطان","كفتان"] ,"fr":["caftan"],"ang":["caftan"]}}' 

			list_clothes = json.loads(Clothes_vals)

			clothes_fr = []
			clothes_ar = []
			clothes_ang = []
			for key in Clothes_keys :
				
				
					
				for i in range(len(list_clothes[key.lower()]["ar"])):

					type_ar = re.search(list_clothes[key.lower()]["ar"][i],msg)
					if type_ar != None :
						clothes_ar.append(key)
						keys_types.append(key)
						group_types.append(type_ar.group(0))
						categories.append("CLOTHES")
						break
				
			
			
			if clothes_ar != "":
				for i in range(len(clothes_ar)):
					
					
					entities_clothes.append(clothes_ar[i])
			
			
			
			entities_clothes = list(dict.fromkeys(entities_clothes))
			
			while("" in entities_clothes) :
				entities_clothes.remove("")
			
			
				

			entity_clothes = self.convert_to_clothes(entities_clothes)
			



			Color_keys = ["BLACK","WHITE","RED","YELLOW","BLUE","GREEN","ORANGE","PINK","BROWN","GREY","PURPLE","BEIGE","COL_VAR"]
			Color_vals = '{"black":{"ar": ["نوار","كحلا","لكحل"] ,"fr":["noir","noire"],"ang":["black"]} , "white" : {"ar": ["ابيض","بلانك","لبيد"] ,"fr":["blanc","blanche"],"ang":["white"]},"red": {"ar": ["روج","حمر","لحمر","لروج"] ,"fr":["rouge"],"ang":[" red "]},"yellow" : {"ar": ["سفر"," جون ","فسفر","لجان"] ,"fr":["jaune"],"ang":["yellow"]},"blue" : {"ar": ["زرق","بلو"] ,"fr":["bleu"],"ang":["blue"]},"green" : {"ar": ["خدر","فرت","لخحدر"] ,"fr":["vert","verte"],"ang":["green"]},"orange" : {"ar":	["ليموني","ارونج","ورانج","لورانج"] ,"fr":["orange"],"ang":["orange"]},"pink" : {"ar": ["روز","وردي","روس"] ,"fr":["rose"],"ang":["pink"]},"brown" : {"ar": ["مارون","بني","لمارون"] ,"fr":["marron"],"ang":["brown"]},"grey" : {"ar": ["رمادي","جري","لجريس"] ,"fr":["gris","grise"],"ang":["grey"]},"purple" : {"ar": ["موف","لماف"] ,"fr":["mauve"],"ang":["purple"]},"beige" : {"ar": ["بيج"] ,"fr":["beige"],"ang":["beige"]}, "col_var" : {"ar": ["لون","كولر"] ,"fr":["couleur"],"ang":["color"]}}' 

			list_Color = json.loads(Color_vals)


			color_fr = []
			color_ar = []
			color_ang = []
			for key in Color_keys :
				
				
					
				for i in range(len(list_Color[key.lower()]["ar"])):

					type_ar = re.search(list_Color[key.lower()]["ar"][i],msg)
					if type_ar != None :
						group_vars.append(type_ar.group(0))
						keys_vars.append(key)
						color_ar.append(key)
						variations.append("COLOR")
						break
				
			
			
			if color_ar != "":
				for i in range(len(color_ar)):
					entities_color.append(color_ar[i])
			
			
			
			entities_color = list(dict.fromkeys(entities_color))
			
			while("" in entities_color) :
				entities_color.remove("")

			entity_color = self.convert_to_color(entities_color)
			


			Size_keys = ["M","S","L","SIZE_VAR"]
			


			Size_vals = '{ "m":{"ar": ["مويان"," م ","مديوم"] ,"fr":["moyen", "moyenne", " m ","medium"],"ang":["medium"]} , "s" : {"ar": ["صغير"," س ","سغيرا","سغير","سمال"] ,"fr":[" s ","small","petit"],"ang":["little"]},"l": {"ar": ["لارج"," ل ","كبير","كبيرا"] ,"fr":[" l ","large"],"ang":["large","big"]},"xs": {"ar": [" اكسترا.سمال "] ,"fr":[" xs ","xsmall"],"ang":["xs","extra.small"]},"size_var": {"ar": ["تايل","حجم"] ,"fr":["taille"],"ang":["size"]}}' 

			list_size = json.loads(Size_vals,strict=False)


			size_fr = []
			size_ar = []
			size_ang = []
			for key in Size_keys :
				
				
					
				for i in range(len(list_size[key.lower()]["ar"])):

					type_ar = re.search(list_size[key.lower()]["ar"][i],msg)
					if type_ar != None :
						size_ar.append(key)
						group_vars.append(type_ar.group(0))
						keys_vars.append(key)
						variations.append("SIZE")
						break
				
			
			
			if size_ar != "":
				for i in range(len(size_ar)):
					
					entities_size.append(size_ar[i])
			
			
			
			entities_size = list(dict.fromkeys(entities_size))
			
			while("" in entities_size) :
				entities_size.remove("")

			entity_size = self.convert_to_size(entities_size)
			



			print("ok1")


			Fabric_keys = ["CHIFFON","COTTON","CREPE","DENIM","LEATHER","SATIN","VELVET","SILK"]
			print("ok2")
			Fabric_vals = '{ "chiffon":{"ar": ["موسلين"] ,"fr":["mousseline"],"ang":["chiffon"]}, "cotton" :{"ar": ["قطن"] ,"fr":["coton"],"ang":["coton"]} ,  "crepe": {"ar": ["كريب"] ,"fr":["crepe"],"ang":["crepe"]},"denim" : {"ar": ["دنيم"] ,"fr":["denim","jean"],"ang":["denim","jeans"]}, "leather" : {"ar": ["كوير","جلد"] ,"fr":["cuir"],"ang":["leather"]},"satin" : {"ar": ["ساتان"] ,"fr":["satin"],"ang":["satin"]},"velvet" : {"ar": ["مخمل","موبرا"] ,"fr":["velour"],"ang":["velvet"]},"silk" : {"ar": ["حرير"] ,"fr":["soie"],"ang":["silk"]} }'
			list_fabric = json.loads(Fabric_vals)
			print("ok3")

			fabric_fr = []
			fabric_ar = []
			fabric_ang = []

			for key in Fabric_keys :
				
				
					
				for i in range(len(list_fabric[key.lower()]["ar"])):

					type_ar = re.search(list_fabric[key.lower()]["ar"][i],msg)
					if type_ar != None :
						fabric_ar.append(key)
						group_vars.append(type_ar.group(0))
						keys_vars.append(key)
						variations.append("FABRIC")
						break
				
			
			
			if fabric_ar != "":
				for i in range(len(fabric_ar)):
					
					
					entities_fabric.append(fabric_ar[i])
			
			
			entities_fabric = list(dict.fromkeys(entities_fabric))
			
			while("" in entities_fabric) :
				entities_fabric.remove("")


		



			entity_Fabric = self.convert_to_fabric(entities_fabric)
			




#Extract Beauty_cosmetics

			entities_cosm = []
			Beauty_cosm_keys = ["SHAMPOING","MASQUE","PARFUM","CREME"]
			print("ok2")
			Beauty_cosm_vals = '{ "shampoing":{"ar": ["شامبوينج"] ,"fr":["shampoing"],"ang":["shampoo"]}, "masque" :{"ar": ["ماسك"] ,"fr":["masque"],"ang":["hair.mask"]}, "parfum" :{"ar": ["بارفوم"] ,"fr":["parfum"],"ang":["perfum"]}, "creme" :{"ar": ["كريم"] ,"fr":["creme"],"ang":["cream"]}}'
			list_cosm = json.loads(Beauty_cosm_vals)
			print("ok3")

			cosm_fr = ''
			cosm_ar = ''
			cosm_ang = ''

			for key in Beauty_cosm_keys :
				
				
					
				for i in range(len(list_cosm[key.lower()]["ar"])):

					type_ar = re.search(list_cosm[key.lower()]["ar"][i],msg)
					if type_ar != None :
						cosm_ar = key
						break
				
			
			
			if cosm_ar != "":

				entities_cosm.append(cosm_ar)
			
			
			
			if len(entities_cosm) == 2 :
				if entities_cosm[0] == entities_cosm[1]:
					entities_cosm[1] = ""
			
			while("" in entities_cosm) :
				entities_cosm.remove("")



			entity_cosm = self.convert_to_cosm(entities_cosm)
			dict_entities_variations = []
			
			if entities_electro != []:
				dict_entities_category.append({"value": entities_electro , "type" :"ELECTRONICS"})
			if entities_access != []:
				dict_entities_category.append({"value": entities_access , "type" :"ACCESSORIES"})
			if entities_clothes != []:
				dict_entities_category.append({"value": entities_clothes , "type" :"CLOTHES"})
			if entities_home_deco != []:
				dict_entities_category.append({"value": entities_home_deco , "type" : "HOME_DECO"})

			
			if entities_size != []:
				dict_entities_variations.append({"value": entities_size , "type" :"SIZE"})
			if entities_color != []:
				dict_entities_variations.append({"value": entities_color , "type" :"COLOR"})
			if entities_fabric != []:
				dict_entities_variations.append({"value": entities_fabric , "type" :"FABRIC"})
			
			if entities_metal != []:
				dict_entities_variations.append({"value": entities_metal , "type" :"METAL"})

			
			print(dict_entities_category)
			print(dict_entities_variations)
			
			types_index = []
			vars_index = []
			group_types = list(dict.fromkeys(group_types))
			group_vars = list(dict.fromkeys(group_vars))

			message.text.lower()

			tokens3 = message.text.lower().split()
			print("tokens3")
			print(tokens3)


			for w in group_types :
				types_index.append(tokens3.index(w))
			for w in group_vars:
				vars_index.append(tokens3.index(w))
			
			
			print("group_types")
			print(group_types)
			print("group_vars")
			print(group_vars)
			print(types_index)
			print(vars_index)

			keys_types = list(dict.fromkeys(keys_types))
			keys_vars = list(dict.fromkeys(keys_vars))



			print("keys_types")
			print(keys_types)
			print("keys_vars")
			print(keys_vars)
			
			
			dict_categories = []
			dict_variations = []
			orders_categories = []
			orders_variations = []
			for i in range (len(group_types)):
				dict_categories.append({"word": group_types[i], "INDEX" : types_index[i]  , "CATEGORY" : keys_types[i] , "TYPE" : categories[i] })
			
			for i in range (len(group_vars)):
				
				dict_variations.append({"word": group_vars[i], "INDEX" : vars_index[i]   , "VARIATION" : keys_vars[i] , "TYPE" : variations[i] })
				
			for k in range(len(dict_categories)):
				orders_categories.append(dict_categories[k]['INDEX'])
			for n in range(len(dict_variations)):
				orders_variations.append(dict_variations[n]['INDEX'])

			print(orders_variations)
			print(orders_categories)
			orders_categories.sort()
			orders_variations.sort()
			print(orders_categories)
			print(orders_variations)
			new_dict_categories = []
			new_dict_variations = []


			for z in range(len(orders_categories)):
				for x in range(len(dict_categories)) :
					if dict_categories[x]["INDEX"] == orders_categories[z] :
						new_dict_categories.append(dict_categories[x])
			
			for z in range(len(orders_variations)):
				for x in range(len(dict_variations)) :
					if dict_variations[x]["INDEX"] == orders_variations[z] :
						new_dict_variations.append(dict_variations[x])

			print(new_dict_variations)
			print(new_dict_categories)

			dict_categories = new_dict_categories
			dict_variations = new_dict_variations

			products = []

			print(dict_variations)
			print(dict_categories)
			if len(dict_categories) < 2  :
				print("< 2")
				
				products.append({ dict_categories[0]['TYPE'] : dict_categories[0]['CATEGORY']})
				for j in range(len(dict_variations)) :
					products[0][dict_variations[j]["TYPE"]] = dict_variations[j]["VARIATION"]
				print(products)
			
			else :
				for i in range(len(dict_categories)) :
                    products.append({ dict_categories[i]['TYPE'] : dict_categories[i]['CATEGORY']}) 
					for j in range(len(dict_variations)) :
						print('ok77')
						print(dict_variations[j])
						print(dict_categories[i])
                        
						if dict_variations[j]['INDEX'] != -1 :
							if dict_variations[j]['INDEX'] < dict_categories[i]['INDEX']   :
								continue
						
										
							else :
								print("apres index")			
								products[i][dict_variations[j]["TYPE"]] = dict_variations[j]["VARIATION"]
								dict_variations[j] = {"word": "", "INDEX" :  -1 , "VARIATION" : "" , "TYPE" : "" }
										
								
						print("products1")
						print(products)

			print("products")
			print(products)
				

									

										



						
						


			

			





			


			
			
			
			

			#dict_entities_category =[{"value": entities_electro , "type" :"ELECTRONICS"},{"value": entities_access , "type" :"ACCESSORIES"},{"value": entities_clothes , "type" :"CLOTHES"},{"value": entities_home_deco , "type" : "HOME_DECO"}]
			#print("dictionary")
			#print(dict_entities_category)
			#dict_entities_variations = []
			#dict_entities_variations =[{"value": entities_size , "type" :"SIZE"},{"value": entities_color , "type" :"COLOR"},{"value": entities_fabric , "type" :"FABRIC"}]
			#print("dictionary2")
			#print(dict_entities_variations)


			entity_category = self.convert_to_category(dict_entities_category)
			message.set("products_categories", entity_category, add_to_output=True)

			entity_variation = self.convert_to_variation(dict_entities_variations)
			message.set("products_variations", entity_variation, add_to_output=True)
			


			
# Extract Numbers 

			numbers = []
			for t in tokens :
				if t.isdigit():
					numbers.append(t)
			
			entity_numb = self.convert_to_number(numbers)
			if numbers != []:

				message.set("Numbers", entity_numb, add_to_output=True)


#Extract amount of money & currencies
			print(tokens)
			amount = ''
			currency_entity = ''
			for s in tokens :
				print(s)
				money = re.search(r'(([0-9]+)+(dh|dhs|mad|euro|eu|€|$|usd))',s)
				if money !=  None :
					print("yes")
					print(money.string)
					for t in money.string :
						if t.isdigit():
							amount = amount + t 
						else :
							currency_entity = currency_entity + t
			

			if amount != None or currency_entity != None :

				currencies_keys = ["MAD","EU","USD"]
				currency = {"MAD":["dh","dhs","mad","dirham","درهم","دم","دحس"],"EU":['euro',"eu","€","اورو"],"USD":["$","usd","دولار"]}
				print(currency["USD"])
				print(tokens)
				for i in range(0,len(tokens)) :
					if tokens[i].isdigit():
						for key in currencies_keys :	
							print(key)
							print("ok3")
							
							print(currency[key])
							if tokens[i+1] in currency[key]  :
								print("ok4")
								amount = tokens[i]
								
								currency_entity = key
								print("currency")
								print(currency)
			
			print("amount")
			print(amount)
			print("currency")
			print(currency_entity)

			

			entity_money = self.convert_to_amount(amount)
			if amount != "" :

				message.set("Amount_Money", entity_money, add_to_output=True)
			
			entity_currency = self.convert_to_currency(currency_entity)

			if currency_entity != "" :

				message.set("Currency", entity_currency, add_to_output=True)

















				
			
			



import os
import pytest

from rasa.nlu import registry, train
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.model import Interpreter, Trainer
from rasa.nlu.training_data import TrainingData
from rasa.utils.tensorflow.constants import EPOCHS
from tests.nlu.conftest import DEFAULT_DATA_PATH


def as_pipeline(*components):
    return [{"name": c, EPOCHS: 1} for c in components]


def pipelines_for_tests():
    # these templates really are just for testing
    # every component should be in here so train-persist-load-use cycle can be
    # tested they still need to be in a useful order - hence we can not simply
    # generate this automatically.

    # Create separate test pipelines for dense featurizers
    # because they can't co-exist in the same pipeline together,
    # as their tokenizers break the incoming message into different number of tokens.

    # first is language followed by list of components
    return [
        ("en", as_pipeline("KeywordIntentClassifier")),
        (
            "en",
            as_pipeline(
                "WhitespaceTokenizer",
                "RegexFeaturizer",
                "LexicalSyntacticFeaturizer",
                "CountVectorsFeaturizer",
                "CRFEntityExtractor",
                "DucklingHTTPExtractor",
                "DIETClassifier",
                "EmbeddingIntentClassifier",
                "ResponseSelector",
                "EntitySynonymMapper",
            ),
        ),
        (
            "en",
            as_pipeline(
                "SpacyNLP",
                "SpacyTokenizer",
                "SpacyFeaturizer",
                "SpacyEntityExtractor",
                "SklearnIntentClassifier",
            ),
        ),
        (
            "en",
            as_pipeline(
                "HFTransformersNLP",
                "LanguageModelTokenizer",
                "LanguageModelFeaturizer",
                "DIETClassifier",
            ),
        ),
        ("en", as_pipeline("ConveRTTokenizer", "ConveRTFeaturizer", "DIETClassifier")),
        (
            "en",
            as_pipeline(
                "MitieNLP", "MitieTokenizer", "MitieFeaturizer", "MitieIntentClassifier"
            ),
        ),
        (
            "zh",
            as_pipeline(
                "MitieNLP", "JiebaTokenizer", "MitieFeaturizer", "MitieEntityExtractor"
            ),
        ),
    ]


def test_all_components_are_in_at_least_one_test_pipeline():
    """There is a template that includes all components to
    test the train-persist-load-use cycle. Ensures that
    really all components are in there."""

    all_components = [c["name"] for _, p in pipelines_for_tests() for c in p]

    for cls in registry.component_classes:
        assert (
            cls.name in all_components
        ), "`all_components` template is missing component."


@pytest.mark.parametrize("language, pipeline", pipelines_for_tests())
async def test_train_persist_load_parse(language, pipeline, component_builder, tmpdir):
    _config = RasaNLUModelConfig({"pipeline": pipeline, "language": language})

    (trained, _, persisted_path) = await train(
        _config,
        path=tmpdir.strpath,
        data=DEFAULT_DATA_PATH,
        component_builder=component_builder,
    )

    assert trained.pipeline

    loaded = Interpreter.load(persisted_path, component_builder)

    assert loaded.pipeline
    assert loaded.parse("Rasa is great!") is not None


@pytest.mark.parametrize("language, pipeline", pipelines_for_tests())
def test_train_model_without_data(language, pipeline, component_builder, tmpdir):
    _config = RasaNLUModelConfig({"pipeline": pipeline, "language": language})

    trainer = Trainer(_config, component_builder)
    trainer.train(TrainingData())
    persisted_path = trainer.persist(tmpdir.strpath)

    loaded = Interpreter.load(persisted_path, component_builder)

    assert loaded.pipeline
    assert loaded.parse("Rasa is great!") is not None


@pytest.mark.parametrize("language, pipeline", pipelines_for_tests())
def test_load_and_persist_without_train(language, pipeline, component_builder, tmpdir):
    _config = RasaNLUModelConfig({"pipeline": pipeline, "language": language})

    trainer = Trainer(_config, component_builder)
    persisted_path = trainer.persist(tmpdir.strpath)

    loaded = Interpreter.load(persisted_path, component_builder)

    assert loaded.pipeline
    assert loaded.parse("Rasa is great!") is not None


async def test_train_model_empty_pipeline(component_builder):
    _config = RasaNLUModelConfig({"pipeline": None, "language": "en"})

    with pytest.raises(ValueError):
        await train(
            _config, data=DEFAULT_DATA_PATH, component_builder=component_builder
        )


async def test_train_named_model(component_builder, tmpdir):
    _config = RasaNLUModelConfig({"pipeline": "keyword", "language": "en"})

    (trained, _, persisted_path) = await train(
        _config,
        path=tmpdir.strpath,
        data=DEFAULT_DATA_PATH,
        component_builder=component_builder,
    )

    assert trained.pipeline

    normalized_path = os.path.dirname(os.path.normpath(persisted_path))
    # should be saved in a dir named after a project
    assert normalized_path == tmpdir.strpath


async def test_handles_pipeline_with_non_existing_component(
    component_builder, pretrained_embeddings_spacy_config
):
    pretrained_embeddings_spacy_config.pipeline.append({"name": "my_made_up_component"})

    with pytest.raises(Exception) as execinfo:
        await train(
            pretrained_embeddings_spacy_config,
            data=DEFAULT_DATA_PATH,
            component_builder=component_builder,
        )
    assert "Cannot find class" in str(execinfo.value)


async def test_train_model_training_data_persisted(component_builder, tmpdir):
    _config = RasaNLUModelConfig({"pipeline": "keyword", "language": "en"})

    (trained, _, persisted_path) = await train(
        _config,
        path=tmpdir.strpath,
        data=DEFAULT_DATA_PATH,
        component_builder=component_builder,
        persist_nlu_training_data=True,
    )

    assert trained.pipeline

    loaded = Interpreter.load(persisted_path, component_builder)

    assert loaded.pipeline
    assert loaded.model_metadata.get("training_data") is not None


async def test_train_model_no_training_data_persisted(component_builder, tmpdir):
    _config = RasaNLUModelConfig({"pipeline": "keyword", "language": "en"})

    (trained, _, persisted_path) = await train(
        _config,
        path=tmpdir.strpath,
        data=DEFAULT_DATA_PATH,
        component_builder=component_builder,
        persist_nlu_training_data=False,
    )

    assert trained.pipeline

    loaded = Interpreter.load(persisted_path, component_builder)

    assert loaded.pipeline
    assert loaded.model_metadata.get("training_data") is None



from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.training_data import Message


def test_spacy_ner_extractor(component_builder, spacy_nlp):
    _config = RasaNLUModelConfig({"pipeline": [{"name": "SpacyEntityExtractor"}]})
    ext = component_builder.create_component(_config.for_component(0), _config)
    example = Message(
        "anywhere in the U.K.",
        {
            "intent": "restaurant_search",
            "entities": [],
            "text_spacy_doc": spacy_nlp("anywhere in the west"),
        },
    )

    ext.process(example, spacy_nlp=spacy_nlp)

    assert len(example.get("entities", [])) == 1
    assert example.get("entities")[0] == {
        "start": 16,
        "extractor": "SpacyEntityExtractor",
        "end": 20,
        "value": "U.K.",
        "entity": "GPE",
        "confidence": None,
    }

    # Test dimension filtering includes only specified dimensions

    example = Message(
        "anywhere in the West with Sebastian Thrun",
        {
            "intent": "example_intent",
            "entities": [],
            "text_spacy_doc": spacy_nlp("anywhere in the West with Sebastian Thrun"),
        },
    )
    _config = RasaNLUModelConfig({"pipeline": [{"name": "SpacyEntityExtractor"}]})

    _config.set_component_attr(0, dimensions=["PERSON"])
    ext = component_builder.create_component(_config.for_component(0), _config)
    ext.process(example, spacy_nlp=spacy_nlp)

    assert len(example.get("entities", [])) == 1
    assert example.get("entities")[0] == {
        "start": 26,
        "extractor": "SpacyEntityExtractor",
        "end": 41,
        "value": "Sebastian Thrun",
        "entity": "PERSON",
        "confidence": None,
    }



import os
import shutil
import tempfile

import pytest

import rasa.data as data
from tests.core.conftest import DEFAULT_STORIES_FILE, DEFAULT_NLU_DATA
from rasa.nlu.training_data import load_data
from rasa.nlu.utils import json_to_string
from rasa.utils import io


def test_get_core_directory(project):
    data_dir = os.path.join(project, "data")
    core_directory = data.get_core_directory([data_dir])
    stories = os.listdir(core_directory)

    assert len(stories) == 1
    assert stories[0].endswith("stories.md")


def test_get_nlu_directory(project):
    data_dir = os.path.join(project, "data")
    nlu_directory = data.get_nlu_directory([data_dir])

    nlu_files = os.listdir(nlu_directory)

    assert len(nlu_files) == 1
    assert nlu_files[0].endswith("nlu.md")


def test_get_nlu_file(project):
    data_file = os.path.join(project, "data/nlu.md")
    nlu_directory = data.get_nlu_directory(data_file)

    nlu_files = os.listdir(nlu_directory)

    original = load_data(data_file)
    copied = load_data(nlu_directory)

    assert nlu_files[0].endswith("nlu.md")
    assert original.intent_examples == copied.intent_examples


def test_get_core_nlu_files(project):
    data_dir = os.path.join(project, "data")
    core_files, nlu_files = data.get_core_nlu_files([data_dir])

    assert len(nlu_files) == 1
    assert list(nlu_files)[0].endswith("nlu.md")

    assert len(core_files) == 1
    assert list(core_files)[0].endswith("stories.md")


def test_get_core_nlu_directories(project):
    data_dir = os.path.join(project, "data")
    core_directory, nlu_directory = data.get_core_nlu_directories([data_dir])

    nlu_files = os.listdir(nlu_directory)

    assert len(nlu_files) == 1
    assert nlu_files[0].endswith("nlu.md")

    stories = os.listdir(core_directory)

    assert len(stories) == 1
    assert stories[0].endswith("stories.md")


def test_get_core_nlu_directories_with_none():
    directories = data.get_core_nlu_directories(None)

    assert all([directory for directory in directories])
    assert all([not os.listdir(directory) for directory in directories])


def test_same_file_names_get_resolved(tmpdir):
    # makes sure the resolution properly handles if there are two files with
    # with the same name in different directories

    tmpdir.join("one").mkdir()
    tmpdir.join("two").mkdir()
    data_dir_one = os.path.join(tmpdir.join("one").join("stories.md").strpath)
    data_dir_two = os.path.join(tmpdir.join("two").join("stories.md").strpath)
    shutil.copy2(DEFAULT_STORIES_FILE, data_dir_one)
    shutil.copy2(DEFAULT_STORIES_FILE, data_dir_two)

    nlu_dir_one = os.path.join(tmpdir.join("one").join("nlu.md").strpath)
    nlu_dir_two = os.path.join(tmpdir.join("two").join("nlu.md").strpath)
    shutil.copy2(DEFAULT_NLU_DATA, nlu_dir_one)
    shutil.copy2(DEFAULT_NLU_DATA, nlu_dir_two)

    core_directory, nlu_directory = data.get_core_nlu_directories([tmpdir.strpath])

    nlu_files = os.listdir(nlu_directory)

    assert len(nlu_files) == 2
    assert all([f.endswith("nlu.md") for f in nlu_files])

    stories = os.listdir(core_directory)

    assert len(stories) == 2
    assert all([f.endswith("stories.md") for f in stories])


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            "dialogflow",
            {
                "data/examples/dialogflow/entities/cuisine.json",
                "data/examples/dialogflow/intents/affirm.json",
                "data/examples/dialogflow/entities/location_entries_es.json",
                "data/examples/dialogflow/intents/affirm_usersays_en.json",
                "data/examples/dialogflow/intents/hi_usersays_es.json",
                "data/examples/dialogflow/entities/cuisine_entries_es.json",
                "data/examples/dialogflow/intents/inform_usersays_en.json",
                "data/examples/dialogflow/intents/hi.json",
                "data/examples/dialogflow/intents/goodbye_usersays_en.json",
                "data/examples/dialogflow/agent.json",
                "data/examples/dialogflow/intents/hi_usersays_en.json",
                "data/examples/dialogflow/entities/location.json",
                "data/examples/dialogflow/intents/affirm_usersays_es.json",
                "data/examples/dialogflow/entities/cuisine_entries_en.json",
                "data/examples/dialogflow/package.json",
                "data/examples/dialogflow/intents/Default Fallback Intent.json",
                "data/examples/dialogflow/intents/goodbye_usersays_es.json",
                "data/examples/dialogflow/intents/goodbye.json",
                "data/examples/dialogflow/entities/location_entries_en.json",
                "data/examples/dialogflow/intents/inform.json",
                "data/examples/dialogflow/intents/inform_usersays_es.json",
            },
        ),
        (
            "luis",
            {
                "data/examples/luis/demo-restaurants_v5.json",
                "data/examples/luis/demo-restaurants_v4.json",
                "data/examples/luis/demo-restaurants_v2.json",
            },
        ),
        (
            "rasa",
            {
                "data/examples/rasa/demo-rasa.json",
                "data/examples/rasa/demo-rasa.md",
                "data/examples/rasa/demo-rasa-responses.md",
                "data/examples/rasa/demo-rasa-multi-intent.md",
            },
        ),
        ("wit", {"data/examples/wit/demo-flights.json"}),
    ],
)
def test_find_nlu_files_with_different_formats(test_input, expected):
    examples_dir = "data/examples"
    data_dir = os.path.join(examples_dir, test_input)
    core_files, nlu_files = data.get_core_nlu_files([data_dir])
    assert nlu_files == expected


def test_is_nlu_file_with_json():
    test = {
        "rasa_nlu_data": {
            "lookup_tables": [
                {"name": "plates", "elements": ["beans", "rice", "tacos", "cheese"]}
            ]
        }
    }

    directory = tempfile.mkdtemp()
    file = os.path.join(directory, "test.json")

    io.write_text_file(json_to_string(test), file)

    assert data.is_nlu_file(file)


def test_is_not_nlu_file_with_json():
    directory = tempfile.mkdtemp()
    file = os.path.join(directory, "test.json")
    io.write_text_file('{"test": "a"}', file)

    assert not data.is_nlu_file(file)



import os
import json
def generateur_dup(direc):
        os.remove("dictionaries/dict_dupont.txt")
        
        mon_fich = open("dictionaries/dict_dupont.txt", "a")
        for fichie in os.listdir(direc):
                
                with open(direc+'/'+fichie, "r",encoding="utf_8") as f:
                        list_dic = json.load(f)
                        
                
                for i in list_dic["rasa_nlu_data"]["common_examples"]:
                        mon_fichier.write(i["text"]+"\n")
                        mon_fich.write(i["text"]+"\n")

        
        
        mon_fich.close()


import argparse
import logging

import rasa.utils.io

from rasa import version
from rasa.cli import (
    scaffold,
    run,
    train,
    interactive,
    shell,
    test,
    visualize,
    data,
    x,
    export,
)
from rasa.cli.arguments.default_arguments import add_logging_options
from rasa.cli.utils import parse_last_positional_argument_as_model_path
from rasa.utils.common import set_log_level
import rasa.utils.tensorflow.environment as tf_env

logger = logging.getLogger(__name__)


def create_argument_parser() -> argparse.ArgumentParser:
    """Parse all the command line arguments for the training script."""

    parser = argparse.ArgumentParser(
        prog="rasa",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Rasa command line interface. Rasa allows you to build "
        "your own conversational assistants 🤖. The 'rasa' command "
        "allows you to easily run most common commands like "
        "creating a new bot, training or evaluating models.",
    )

    parser.add_argument(
        "--version",
        action="store_true",
        default=argparse.SUPPRESS,
        help="Print installed Rasa version",
    )

    parent_parser = argparse.ArgumentParser(add_help=False)
    add_logging_options(parent_parser)
    parent_parsers = [parent_parser]

    subparsers = parser.add_subparsers(help="Rasa commands")

    scaffold.add_subparser(subparsers, parents=parent_parsers)
    run.add_subparser(subparsers, parents=parent_parsers)
    shell.add_subparser(subparsers, parents=parent_parsers)
    train.add_subparser(subparsers, parents=parent_parsers)
    interactive.add_subparser(subparsers, parents=parent_parsers)
    test.add_subparser(subparsers, parents=parent_parsers)
    visualize.add_subparser(subparsers, parents=parent_parsers)
    data.add_subparser(subparsers, parents=parent_parsers)
    export.add_subparser(subparsers, parents=parent_parsers)
    x.add_subparser(subparsers, parents=parent_parsers)

    return parser


def print_version() -> None:
    print("Rasa", version.__version__)


def main() -> None:
    # Running as standalone python application
    import os
    import sys

    parse_last_positional_argument_as_model_path()
    arg_parser = create_argument_parser()
    cmdline_arguments = arg_parser.parse_args()

    log_level = (
        cmdline_arguments.loglevel if hasattr(cmdline_arguments, "loglevel") else None
    )
    set_log_level(log_level)

    tf_env.setup_tf_environment()

    # insert current path in syspath so custom modules are found
    sys.path.insert(1, os.getcwd())

    if hasattr(cmdline_arguments, "func"):
        rasa.utils.io.configure_colored_logging(log_level)
        cmdline_arguments.func(cmdline_arguments)
    elif hasattr(cmdline_arguments, "version"):
        print_version()
    else:
        # user has not provided a subcommand, let's print the help
        logger.error("No command specified.")
        arg_parser.print_help()
        exit(1)


if __name__ == "__main__":
    main()



import re
from typing import Any, Dict, List, Text, Optional
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.components import Component
from rasa.nlu.tokenizers.tokenizer_fr import Token, Tokenizer
from rasa.nlu.training_data import TrainingData, Message
from itertools import *
from custom_func.generateur_dup import generateur_dup
from rasa.nlu.constants import TOKENS_NAMES, MESSAGE_ATTRIBUTES, TEXT,INTENT
import jellyfish



class WhitespaceTokenizer_dup(Tokenizer):
    
    defaults = {
        # Flag to check whether to split intents
        "intent_tokenization_flag": False,
        # Symbol on which intent should be split
        "intent_split_symbol": "_",
        # Text will be tokenized with case sensitive as default
        "case_sensitive": True,
    }

    def __init__(self, component_config: Dict[Text, Any] = None) -> None:
        """Construct a new tokenizer using the WhitespaceTokenizer framework."""

        super().__init__(component_config)

        self.case_sensitive = self.component_config["case_sensitive"]

    def tokenize(self, message: Message, attribute: Text) -> List[Token]:
        text = message.get(attribute)

        
        text = text.lower()
        
        numb = []

        toks = ''
        toks = text.split()
        numbs = []
        
        
        stop_words = ['en','homme','femme','avez','qu','et','ceci','cela','si','pouvez','tout','vous','sur','dans','peut','peux','votre','c','quoi','t','pouvez','pour','me','veuillez','s','d','ca','cette','ce','est','il','a','plus','est-il','avoir','etre','que','qui','si','veux','en','j','ai','des','les','quelles','sont','quels','du','que','ce','est','je','un','une','le','la','de','avec','sur','moi','avec','des','stp','plait','les','tu','puis','svp','par','aux','et','voudrais','souhaite','souhaiterai','me','d','puis','m','mes']


        for s in toks:
            if s.isdigit():
               numb.append(s)
        
        for i in range(0,len(numb)):

            text = text+" "+ str(numb[i]) + " " 
        
        for i in range(0,len(numbs)):

            text = text+" "+ str(numbs[i]) + " " 
        
        # remove 'not a word character' if
        print(text)

        # remove 'not a word character' if
        words = re.sub(
            # there is a space or an end of a string after it
            r"[^\w#@&]+(?=\s|$)|"
            # there is a space or beginning of a string before it
            # not followed by a number
            r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
            # not in between numbers and not . or @ or & or - or #
            # e.g. 10'000.00 or blabla@gmail.com
            # and not url characters
            r"(?<=[^0-9\s])[^\w._~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
            " ",
            text,
        ).split()
        # if we removed everything like smiles `:)`, use the whole text as 1 token
        if not words:
            words = [text]
        with open("dictionaries/dict_dupont.txt", 'r',encoding="utf8") as f:
            list = set(chain(*(line.split() for line in f if line)))
        

        dicto = {}
        # if we removed everything like smiles `:)`, use the whole text as 1 token
        #words2 = []
        
        tokens = []
        running_offset = 0
        for word in words:
            

            
            if word.isnumeric() :
                min_dist = word
            else:   
                for mot in list:
                    dicto[mot]=jellyfish.levenshtein_distance(mot, word)
                min_dist = min(dicto.keys(), key=(lambda k: dicto[k]))
            
            if jellyfish.levenshtein_distance(word,min_dist) > 1:
                min_dist = word
            
            
            if min_dist in stop_words:
                continue
            else :
                word_corr = min_dist
                print("min"+min_dist)
                word_offset = text.index(word, running_offset)
                word_len = len(word)
                running_offset = word_offset + word_len
                tokens.append(Token(word_corr, word_offset))

        return tokens
    def tokenize_tr(self, message: Message, attribute: Text) -> List[Token]:
        text = message.get(attribute)

        stop_words = ['en','homme','femme','avez','qu','et','ceci','cela','si','pouvez','tout','avant','vous','sur','dans','peut','peux','votre','c','quoi','t','pouvez','pour','me','veuillez','s','d','ca','cette','ce','est','il','a','plus','est-il','avoir','etre','que','qui','si','veux','en','j','ai','des','les','quelles','sont','quels','du','que','ce','est','je','un','une','le','la','de','avec','sur','moi','avec','des','stp','plait','les','tu','puis','svp','par','aux','et','voudrais','souhaite','souhaiterai','me','d','puis','m','avant','mes']
        text = text.lower()
        

        # remove 'not a word character' if
        words = re.sub(
            # there is a space or an end of a string after it
            r"[^\w#@&]+(?=\s|$)|"
            # there is a space or beginning of a string before it
            # not followed by a number
            r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
            # not in between numbers and not . or @ or & or - or #
            # e.g. 10'000.00 or blabla@gmail.com
            # and not url characters
            r"(?<=[^0-9\s])[^\w._~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
            " ",
            text,
        ).split()
        # if we removed everything like smiles `:)`, use the whole text as 1 token
        words2 =[]
        

        for word in words :
            if word not in stop_words:
                words2.append(word)
                print("appended")
            else :
                print('stop word')
        print(words2)
        # if we removed everything like smiles `:)`, use the whole text as 1 token
        if not words2:
            words2 = [text]
        print("ok")
        
        
        tokens = []
        running_offset = 0
        for word in words2:
            word_offset = text.index(word, running_offset)
            word_len = len(word)
            running_offset = word_offset + word_len
            tokens.append(Token(word, word_offset))

        return tokens
    
    def process(self, message: Message, **kwargs: Any) -> None:
        """Tokenize the incoming message."""
        tokens = self.tokenize(message, TEXT)
        tokens = self.add_cls_token(tokens, TEXT)
        message.set(TOKENS_NAMES[TEXT], tokens)
    
    def train(
        self,
        training_data: TrainingData,
        config: Optional[RasaNLUModelConfig] = None,
        **kwargs: Any,
    ) -> None:
        generateur_dup('./data_dup/data')
        """Tokenize all training data."""

        for example in training_data.training_examples:
            for attribute in MESSAGE_ATTRIBUTES:
                if example.get(attribute) is not None:
                    if attribute == INTENT:
                        tokens = self._split_intent(example)
                    else:
                        tokens = self.tokenize_tr(example, attribute)
                        tokens = self.add_cls_token(tokens, attribute)
                    example.set(TOKENS_NAMES[attribute], tokens)



import pytest
from typing import Text, Dict
from rasa.utils.tensorflow.environment import _parse_gpu_config


@pytest.mark.parametrize(
    "gpu_config_string, parsed_gpu_config",
    [("0: 1024", {0: 1024}), ("0:1024, 1:2048", {0: 1024, 1: 2048})],
)
def test_gpu_config_parser(gpu_config_string: Text, parsed_gpu_config: Dict[int, int]):
    assert _parse_gpu_config(gpu_config_string) == parsed_gpu_config



import argparse
from pathlib import Path
from typing import Callable, Optional, Dict, Text, List, Tuple, Any
from unittest.mock import Mock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from _pytest.pytester import RunResult

import rasa.core.utils as rasa_core_utils
from rasa.cli import export
from rasa.core.brokers.pika import PikaEventBroker
from rasa.core.events import UserUttered
from rasa.core.trackers import DialogueStateTracker
from rasa.exceptions import PublishingError, NoEventsToMigrateError
from tests.conftest import (
    MockExporter,
    random_user_uttered_event,
    write_endpoint_config_to_yaml,
)


def test_export_help(run: Callable[..., RunResult]):
    output = run("export", "--help")

    help_text = """usage: rasa export [-h] [-v] [-vv] [--quiet] [--endpoints ENDPOINTS]
                   [--minimum-timestamp MINIMUM_TIMESTAMP]
                   [--maximum-timestamp MAXIMUM_TIMESTAMP]
                   [--conversation-ids CONVERSATION_IDS]"""

    lines = help_text.split("\n")

    for i, line in enumerate(lines):
        assert output.outlines[i] == line


@pytest.mark.parametrize(
    "minimum_timestamp,maximum_timestamp",
    [(2, 3), (None, 5.5), (None, None), (5, None)],
)
def test_validate_timestamp_options(
    minimum_timestamp: Optional[float], maximum_timestamp: Optional[float],
):
    args = argparse.Namespace()
    args.minimum_timestamp = (
        str(minimum_timestamp) if minimum_timestamp is not None else None
    )
    args.maximum_timestamp = (
        str(maximum_timestamp) if maximum_timestamp is not None else None
    )

    # no error is raised
    # noinspection PyProtectedMember
    export._assert_max_timestamp_is_greater_than_min_timestamp(args)


def test_validate_timestamp_options_with_invalid_timestamps():
    args = argparse.Namespace(minimum_timestamp=3, maximum_timestamp=2)
    with pytest.raises(SystemExit):
        # noinspection PyProtectedMember
        export._assert_max_timestamp_is_greater_than_min_timestamp(args)


# noinspection PyProtectedMember
def test_get_event_broker_and_tracker_store_from_endpoint_config(tmp_path: Path):
    # write valid config to file
    endpoints_path = write_endpoint_config_to_yaml(
        tmp_path, {"event_broker": {"type": "sql"}, "tracker_store": {"type": "sql"}}
    )

    available_endpoints = rasa_core_utils.read_endpoints_from_path(endpoints_path)

    # fetching the event broker is successful
    assert export._get_event_broker(available_endpoints)
    assert export._get_tracker_store(available_endpoints)


# noinspection PyProtectedMember
def test_get_event_broker_from_endpoint_config_error_exit(tmp_path: Path):
    # write config without event broker to file
    endpoints_path = write_endpoint_config_to_yaml(
        tmp_path, {"tracker_store": {"type": "sql"}}
    )

    available_endpoints = rasa_core_utils.read_endpoints_from_path(endpoints_path)

    with pytest.raises(SystemExit):
        assert export._get_event_broker(available_endpoints)


def test_get_tracker_store_from_endpoint_config_error_exit(tmp_path: Path):
    # write config without event broker to file
    endpoints_path = write_endpoint_config_to_yaml(tmp_path, {})

    available_endpoints = rasa_core_utils.read_endpoints_from_path(endpoints_path)

    with pytest.raises(SystemExit):
        # noinspection PyProtectedMember
        assert export._get_tracker_store(available_endpoints)


@pytest.mark.parametrize(
    "requested_ids,expected",
    [("id1", ["id1"]), ("id1,id2", ["id1", "id2"]), (None, None), ("", None)],
)
def test_get_requested_conversation_ids(
    requested_ids: Optional[Text], expected: Optional[List[Text]]
):
    # noinspection PyProtectedMember
    assert export._get_requested_conversation_ids(requested_ids) == expected


def test_prepare_pika_event_broker():
    # mock a pika event broker
    pika_broker = Mock(spec=PikaEventBroker)

    # patch the spinner so we can execute the `_prepare_pika_producer()` function
    pika_broker.is_ready.return_value = True

    # noinspection PyProtectedMember
    export._prepare_event_broker(pika_broker)

    # the attributes are set as expected
    assert not pika_broker.should_keep_unpublished_messages
    assert pika_broker.raise_on_failure


@pytest.mark.parametrize(
    "current_timestamp,maximum_timestamp,endpoints_path,requested_ids,expected",
    [
        (1.0, None, None, None, "--minimum-timestamp 1.0"),
        (1.0, None, None, ["5", "6"], "--minimum-timestamp 1.0 --conversation-ids 5,6"),
        (1.0, 3.4, None, None, "--minimum-timestamp 1.0 --maximum-timestamp 3.4"),
        (
            1.0,
            2.5,
            "a.yml",
            None,
            "--endpoints a.yml --minimum-timestamp 1.0 --maximum-timestamp 2.5",
        ),
        (
            1.0,
            2.5,
            "a.yml",
            ["1", "2", "3"],
            (
                "--endpoints a.yml --minimum-timestamp 1.0 --maximum-timestamp 2.5 "
                "--conversation-ids 1,2,3"
            ),
        ),
    ],
)
def test_get_continuation_command(
    current_timestamp: float,
    maximum_timestamp: Optional[float],
    endpoints_path: Optional[Text],
    requested_ids: Optional[List[Text]],
    expected: Text,
):
    exporter = MockExporter()
    exporter.maximum_timestamp = maximum_timestamp
    exporter.endpoints_path = endpoints_path
    exporter.requested_conversation_ids = requested_ids

    # noinspection PyProtectedMember
    assert (
        export._get_continuation_command(exporter, current_timestamp)
        == f"rasa export {expected}"
    )


def _add_conversation_id_to_event(event: Dict, conversation_id: Text):
    event["sender_id"] = conversation_id


def prepare_namespace_and_mocked_tracker_store_with_events(
    temporary_path: Path, monkeypatch: MonkeyPatch
) -> Tuple[List[UserUttered], argparse.Namespace]:
    endpoints_path = write_endpoint_config_to_yaml(
        temporary_path,
        {"event_broker": {"type": "pika"}, "tracker_store": {"type": "sql"}},
    )

    # export these conversation IDs
    all_conversation_ids = ["id-1", "id-2", "id-3"]

    requested_conversation_ids = ["id-1", "id-2"]

    # create namespace with a set of cmdline arguments
    namespace = argparse.Namespace(
        endpoints=endpoints_path,
        conversation_ids=",".join(requested_conversation_ids),
        minimum_timestamp=1.0,
        maximum_timestamp=10.0,
    )

    # prepare events from different senders and different timestamps
    events = [random_user_uttered_event(timestamp) for timestamp in [1, 2, 3, 4, 11, 5]]
    events_for_conversation_id = {
        all_conversation_ids[0]: [events[0], events[1]],
        all_conversation_ids[1]: [events[2], events[3], events[4]],
        all_conversation_ids[2]: [events[5]],
    }

    def _get_tracker(conversation_id: Text) -> DialogueStateTracker:
        return DialogueStateTracker.from_events(
            conversation_id, events_for_conversation_id[conversation_id]
        )

    # mock tracker store
    tracker_store = Mock()
    tracker_store.keys.return_value = all_conversation_ids
    tracker_store.retrieve.side_effect = _get_tracker

    monkeypatch.setattr(export, "_get_tracker_store", lambda _: tracker_store)

    return events, namespace


def test_export_trackers(tmp_path: Path, monkeypatch: MonkeyPatch):
    events, namespace = prepare_namespace_and_mocked_tracker_store_with_events(
        tmp_path, monkeypatch
    )

    # mock event broker so we can check its `publish` method is called
    event_broker = Mock()
    event_broker.publish = Mock()
    monkeypatch.setattr(export, "_get_event_broker", lambda _: event_broker)

    # run the export function
    export.export_trackers(namespace)

    # check that only events 1, 2, 3, and 4 have been published
    # event 6 was sent by `id-3` which was not requested, and event 5
    # lies outside the requested time range
    calls = event_broker.publish.mock_calls

    # only four events were published (i.e. `publish()` method was called four times)
    assert len(calls) == 4

    # call objects are tuples of (name, pos. args, kwargs)
    # args itself is a tuple, and we want to access the first one, hence `call[1][0]`
    # check that events 1-4 were published
    assert all(
        any(call[1][0]["text"] == event.text for call in calls) for event in events[:4]
    )


@pytest.mark.parametrize("exception", [NoEventsToMigrateError, PublishingError(123)])
def test_export_trackers_publishing_exceptions(
    tmp_path: Path, monkeypatch: MonkeyPatch, exception: Exception
):
    events, namespace = prepare_namespace_and_mocked_tracker_store_with_events(
        tmp_path, monkeypatch
    )

    # mock event broker so we can check its `publish` method is called
    event_broker = Mock()
    event_broker.publish.side_effect = exception
    monkeypatch.setattr(export, "_get_event_broker", lambda _: event_broker)

    with pytest.raises(SystemExit):
        export.export_trackers(namespace)
